(* This module contains data types and functions necessary for parsing the 
 * CVC input syntax (Stanford's Cooperative Validity Checker), and for 
 * generating the Input data structures used by the rest of the prover *)
module I = Input
module L = Logic
module H = Hashtbl

module E = Errormsg
open Pretty


let preprocess = ref false

type typ = 
    TId of string           (* A declared type *)
  | TReal 
  | TBool 
  | TTuple of typ list
  | TRecord of (string * typ) list
  | TEnum of string list (* An enumeration type *)
  | TFun of (typ * typ)
  | TArray of typ * typ

type expr = 
    EId of string
  | ENId of string (* A negated boolean constant *)

  | ENum of int
  | EBool of bool

  | EPlus of expr * expr
  | EMinus of expr * expr
  | EMult of expr * expr
  
  | EEq of expr * expr
  | ENeq of expr * expr
  | EGeq of expr * expr
  | ELeq of expr * expr
  | EGt of expr * expr
  | ELt of expr * expr

  | ENot of expr
  | EAnd of expr * expr
  | EOr of expr * expr
  | EXor of expr * expr
  | EImpl of expr * expr
  | EIff of expr * expr

  | EIf of expr * expr * expr
  | ELet of (string * typ * expr) list * expr

  | ETuple of expr list
  | EApp   of string * expr list 
  | ENApp  of string * expr list (* A negated boolean predicate *)

  | EAll of string * typ * expr (* Universal quantification *)
  | EExist of string * typ * expr (* Existential quantification *)

  | EUpd of expr * expr list * expr (* Update in a multi-dimensional array *)
  | ESel of expr * expr (* Select *)

type input = 
    ConstDecl of string list * typ
  | ConstInitDecl of string * typ * expr (* A constant with initializer *)
  | TypeConstDecl of string list
  | TypeConstInitDecl of string * typ
  | Query of expr
  | Assert of expr
  | Retract
  | Print of expr 
  | Echo of string
  | Block of (string * string) list
  | EndBlock 
  | Eof  


let rec d_typ () = function
    TId n -> text n
  | TReal -> text "REAL"
  | TBool -> text "BOOL"
  | TTuple tl -> 
      dprintf "[@[%a@]]" (docList ~sep:(chr ',' ++ break) (d_typ ())) tl
  | TRecord tl -> 
      dprintf "[#@[%a@]#]" 
        (docList ~sep:(chr ',' ++ break) 
           (fun (n, t) -> dprintf "%s : %a" n d_typ t)) tl
  | TFun (t1, t2) -> dprintf "[%a -> %a]" d_typ t1 d_typ t2
  | TEnum nl -> dprintf "{@[%a@]}" (docList ~sep:(chr ',' ++ break) text) nl
  | TArray (tidx, telem) -> dprintf "ARRAY(%a -> %a)" 
        d_typ tidx d_typ telem
        

let rec d_exp () (e: expr) = 
  let d_binop (op: string) (e1: expr) (e2: expr) = 
    dprintf "(@[%a %s@? %a@])" d_exp e1 op d_exp e2
  in
  match e with 
    EId n -> text n
  | ENId n -> text "NOT n"
  | EBool b -> dprintf "%b" b
  | ENum n -> num n
  | ENot e -> dprintf "(NOT %a)" d_exp e
  | EPlus (e1, e2) -> d_binop "+" e1 e2
  | EMinus (e1, e2) -> d_binop "-" e1 e2
  | EMult (e1, e2) -> d_binop "*" e1 e2
  | EEq (e1, e2) -> d_binop "=" e1 e2
  | ENeq (e1, e2) -> d_binop "<>" e1 e2
  | EGeq (e1, e2) -> d_binop ">=" e1 e2
  | EGt (e1, e2) -> d_binop ">" e1 e2
  | ELeq (e1, e2) -> d_binop "<=" e1 e2
  | ELt (e1, e2) -> d_binop "<" e1 e2

  | EAnd (e1, e2) -> d_binop "AND" e1 e2
  | EOr (e1, e2) -> d_binop "OR" e1 e2
  | EXor (e1, e2) -> d_binop "XOR" e1 e2

  | EImpl (e1, e2) -> d_binop "IMPL" e1 e2
  | EIff (e1, e2) -> d_binop "IFF" e1 e2
        
  | EIf (b, et, ef) -> 
      dprintf "(@[if %a@?THEN %a@?ELSE %a@])" d_exp b d_exp et d_exp ef
  | EUpd (a, idx, v) -> 
      dprintf "(%a WITH %a := %a)"
        d_exp a
        (docList ~sep:nil (fun i -> dprintf "[%a]" d_exp i)) idx
        d_exp v
  | ESel (a, idx) -> dprintf "%a[%a]" d_exp a d_exp idx

  | ELet (nl, body) -> 
      dprintf "(@[let @[%a@]@?in %a@])" 
        (docList ~sep:(text "," ++ break) 
           (fun (n, t, e) -> dprintf "%s : %a = %a" n d_typ t d_exp e))
        nl
        d_exp body

  | EApp (f, args) -> 
      dprintf "%s%a" f d_exp (ETuple args)
  | ENApp (f, args) -> 
      dprintf "NOT %s%a" f d_exp (ETuple args)
  | ETuple exprs -> 
      dprintf "(@[%a@])"
        (docList ~sep:(chr ',' ++ break) (d_exp ())) exprs
  | EAll (x, t, body) -> 
      dprintf "(FOR@[ALL %s:%a.%a@])"
        x d_typ t d_exp body
  | EExist (x, t, body) -> 
      dprintf "(EXI@[STS %s:%a.%a@])"
        x d_typ t d_exp body

(* Count the queries here *)
let nextQueryId = ref 0

(* Keep here the constants *)
let constantDecls: (string, typ) H.t = H.create 113
let typeConstantDecls: (string, unit) H.t = H.create 113

(* Keep here the global names *)
let names: (string, typ * expr) H.t = H.create 113
let typeNames: (string, typ) H.t = H.create 113


(* Convert an arithmetic expression *)
let rec convArith (e: expr) : L.exp = 
  match e with 
    ENum n -> L.integer n
  | EId n -> begin
      (* See if this is a constant or not *)
      try
        let _, e = H.find names n in
        convArith e
      with Not_found -> begin
        try
          let _ = H.find constantDecls n in
          L.Fun(n, []) (* A constant *)
        with Not_found -> 
          L.NVar n (* Must be a variable *)
      end
  end
  | EPlus (e1, e2) -> L.plus (convArith e1) (convArith e2)
  | EMinus (e1, e2) -> L.minus (convArith e1) (convArith e2)
  | EMult (e1, e2) -> L.times (convArith e1) (convArith e2)
  | ELet (nl, body) -> 
      List.iter (fun (n, t, e) -> H.add names n (t, e)) nl;
      let res = convArith body in
      List.iter (fun (n, _, _) -> H.remove names n) nl;
      res
  | EIf _ -> 
      E.s (E.unimp "If in convArith")

  | EApp(n, args) -> L.Fun(n, List.map convArith args)

  | ESel (a, idx) -> 
      L.Fun("sel", [ convArith idx; convArith a ])

  | EUpd (a, idx, v) -> 
      (* The last index is the one being updated. The previous ones are used 
       * to find the array being updated *)
      let rec loop (arr: L.exp) = function
          [ updidx ] -> arr, convArith updidx
        | selidx :: rest -> 
            loop (L.Fun("sel", [ convArith selidx; arr ])) rest
        | [] -> E.s (E.bug "convArith: empty indices")
      in
      let arr, updidx = loop (convArith a) idx in
      L.Fun("upd", [updidx; convArith v; arr])

  | e -> E.s (E.unimp "convArith: %a" d_exp e)


(* Convert a predicate *)
let rec convPred (p: expr) : L.pred = 
  match p with 
  | EImpl (p1, p2) -> L.Imply (convPred p1, convPred p2)
  | EAnd (p1, p2) -> L.And [convPred p1; convPred p2]
  | EEq (e1, e2) -> L.eq (convArith e1) (convArith e2)
  | ENeq (e1, e2) -> L.ne (convArith e1) (convArith e2)
  | EAll (var, _, p) -> L.All (var, convPred p)

  | EId n -> begin (* This name must be a bool *)
      (* See if it is a name *)
      try
        let _, e = H.find names n in
        convPred e
      with Not_found -> 
        (* Must be a constant *)
        L.eq (L.NVar n) L.one
  end
  | ENId n -> 
      (* Must be a constant *)
        L.eq (L.NVar n) L.zero

  | EApp _ -> 
      L.eq (convArith p) L.one
  | ENApp _ -> 
      L.eq (convArith p) L.zero
      

  | EBool true -> L.Bool true
  | EBool false -> L.Bool false

  | EGeq (e1, e2) -> 
      L.geq (convArith e1) (convArith e2)
  | EGt  (e1, e2) -> 
      L.gt (convArith e1) (convArith e2)
  | ELeq (e1, e2) -> 
      L.leq (convArith e1) (convArith e2)
  | ELt  (e1, e2) -> 
      L.lt (convArith e1) (convArith e2)


  | EIff (e1, e2) ->
      let e1' = convPred e1 in
      let e2' = convPred e2 in
      L.And [L.Imply(e1', e2'); L.Imply(e2', e1') ]

  | EOr (e1, e2) -> L.Or [convPred e1;  convPred e2]

  | ELet (nl, body) -> 
      List.iter (fun (n, t, e) -> H.add names n (t, e)) nl;
      let res = convPred body in
      List.iter (fun (n, _, _) -> H.remove names n) nl;
      res

  | EIf (b, et, ef) ->
      L.And [L.Imply(convPred b, convPred et);
             L.Imply(convPred (ENot b), convPred ef) ]

  | ENot(p) -> (L.negatePred (convPred p))

  | _ -> E.s (E.unimp "convPred: %a" d_exp p)

    
(* Remember here the assert statements *)
let open_asserts: int ref = ref 0
let open_blocks: int ref = ref 0

let nextGoalName () = 
  let gname = "Query" ^ (string_of_int !nextQueryId) in
  incr nextQueryId;
  gname
  
let rec parseNextInput (theparser: unit -> input) : I.input list = 
  let res = theparser () in
  Parsing.clear_parser ();
  (* Keep parsing until we see a query *)
  match res with
    Eof ->  [ I.Eof ]

  | ConstDecl (clst, t) -> 
      List.iter (fun n -> H.add constantDecls n t) clst;
      parseNextInput theparser

  | ConstInitDecl (n, t, e) -> 
      H.add names n (t, e);
      parseNextInput theparser

  | TypeConstDecl (clst) -> 
      List.iter (fun n -> H.add typeConstantDecls n ()) clst;
      parseNextInput theparser

  | TypeConstInitDecl (n, t) -> 
      H.add typeNames n t;
      parseNextInput theparser
 
  | Print e -> ignore (E.log "%a" d_exp e); parseNextInput theparser

  | Echo s -> 
      ignore (E.log "%s" s); 
      parseNextInput theparser
 
  | Query e -> 
      (* If we are at top-level, start a new goal *)
      if !open_asserts = 0 then begin
        [ I.StartGoal (nextGoalName ()); I.Prove (convPred e); I.EndGoal ]
      end else 
        [ I.Prove (convPred e) ]

  | Assert e ->
      (* If this is a top-level assert, start a new goal *)
      incr open_asserts;
      if !open_asserts = 1 && !open_blocks = 0 then begin
        [ I.StartGoal (nextGoalName ()); I.Assume (convPred e) ]
      end else 
        [ I.Assume (convPred e) ]

  | Retract -> begin (* Retract the last assert *)
      if !open_asserts = 0 then 
        E.s (E.error "Found retract but with no asserts");
      decr open_asserts;
      if !open_asserts = 0 then 
        [ I.UnAssume; I.EndGoal ]
      else
        [ I.UnAssume ]
  end
  | Block pars -> 
      incr open_blocks;
      if !open_blocks = 1 && !open_asserts = 0 then begin
        [ I.StartGoal (nextGoalName ()); I.Block pars ]
      end else 
        [ I.Block pars ]

  | EndBlock ->
      if !open_blocks = 0 then 
        E.s (E.error "Found EndBlock but with no Block");
      decr open_blocks;
      [ I.UnBlock ]
      
        
let initParser () = 
  nextQueryId := 0;
  H.clear constantDecls;
  H.clear names;
  ()
  

