{
module E = Errormsg
module S = Satproc

open Cvcparse

(*** Error handling ***)
let lexerror msg =
  E.parse_error msg


let scan_escape str =
  match str with
    "n" -> "\n"
  | "r" -> "\r"
  | "t" -> "\t"
  | "b" -> "\b"
  | "f" -> "\012"  (* ASCII code 12 *)
  | "v" -> "\011"  (* ASCII code 11 *)
  | "a" -> "\007"  (* ASCII code 7 *)
  | "e" -> "\027"  (* ASCII code 27. This is a GCC extension *)
  | _ -> str


}

let digit    = ['0'-'9']
let letter   = ['a'-'z' 'A'-'Z' '_' '$' '@'] 
let letdig   = ['0'-'9' 'a'-'z' 'A'-'Z' '_' '$' '@' ]

let capital  = ['A'-'Z']
let small    = ['a'-'z' '$' '_']
let ws       = [' ' '\009' '\012']
let pathname = ['a'-'z' 'A'-'Z' '0'-'9' '.' '/' '\\' '-']

let escape = '\\' _

rule token = parse
    '\n'		{ begin
			    E.newline ();
			    token lexbuf 
			  end }

| '%'                   { endline lexbuf }

| ws			{ token lexbuf }
| "=>"                  { IMPL }
| "<=>"                 { IFF }
| "/="                  { NEQ }
| "="                   { EQ }
| "<"			{ LT }
| ">"			{ GT }
| "<="		        { LEQ }
| ">="		        { GEQ }
| "->"                  { ARROW }
| "[#"                  { SQHASH }
| "#]"                  { HASHSQ }
| ":="			{ ASSIGN }
| '+'			{ PLUS }
| '-'			{ MINUS }
| '*'			{ MULT }
| ':'			{ COLON }
| ';'                   { SEMICOLON }
| ','                   { COMMA }
| '('			{ LPAREN }
| ')'			{ RPAREN }
| '['			{ LBRACKET }
| ']'			{ RBRACKET }
| '{'			{ LBRACE }
| '}'			{ RBRACE }
| '.'                   { DOT }

| "AND"		      { AND }
| "ARRAY"             { ARRAY }
| "BOOLEAN"	      { BOOLEAN }
| "BLOCK"             { BLOCK }
| "ENDBLOCK"          { ENDBLOCK }
| "DATATYPE"	      { DATATYPE }
| "ELSE" 	      { ELSE }
| "ELSIF"  	      { ELSIF }
| "ENDIF"             { ENDIF }
| "END"		      { END }
| "EXISTS"            { EXISTS }
| "FALSE"	      { FALSE }
| "FORALL"	      { ALL }
| "IF"		{ IF }
| "IN"		{ IN }
| "LAMBDA"	{ LAMBDA }
| "LET"		{ LET }
| "NOT"		{ NOT }
| "OF"		{ OF }
| "OR"		{ OR }
| "REAL"		{ REAL }
| "THEN"		{ THEN }
| "TRUE"		{ TRUE }
| "TYPE"		{ TYPE }
| "WITH"		{ WITH }
| "XOR"		{ XOR }
| "TCC"		{ TCC }

| "ASSERT"	{ ASSERT }
| "RETRACT"     { RETRACT }

| "QUERY"	        { QUERY }
| "DBG"	        { DBG }
| "TRACE"	        { TRACE }
| "OPTION"        { OPTION }
| "HELP"	        { HELP }
| "TRANSFORM"	{ TRANSFORM }
| "PRINT"	        { PRINT }
| "CALL"	        { CALL }
| "ECHO"	        { ECHO }
| "INCLUDE"       { INCLUDE }
| "DUMP_ASSUMPTIONS"   { DUMP_ASSUMPTIONS }
| "DUMP_PROOF"    { DUMP_PROOF }
| "DUMP_SIG"      { DUMP_SIG }
| "WHERE"         { WHERE }
| "PUSH"          { PUSH }
| "POP"           { POP }
| "POPTO"         { POPTO }
| "RESET"         { RESET }
| "CONTEXT"       { CONTEXT }
| "FORGET"        { FORGET }
| "GET_TYPE"      { GET_TYPE }
| "CHECK_TYPE"    { CHECK_TYPE }
| "GET_CHILD"     { GET_CHILD }
| "GET_OP"        { GET_OP }
| "SUBSTITUTE"    { SUBSTITUTE }

| (digit)+	{ let str = Lexing.lexeme lexbuf in
                  let len = String.length str in
		  let zero = Char.code '0' in
                  assert (len > 0);
		  let rec accum a d =
		    let acc c = a + (d * ((Char.code c) - zero)) in
		    function
			0 -> let c = str.[0] in
			if c='~' then - a else (acc c)
		      | i -> accum (acc str.[i]) (d * 10) (i - 1)
		  in
		  Num (accum 0 1 (len-1)) }

| (letter)letdig*	{ Id (Lexing.lexeme lexbuf) }
| '"'			{ (* '"' *)
                            try STRING (str lexbuf)
                            with e -> 
                                raise (Failure 
                                        ("str: " ^ Printexc.to_string e)) 
                        } 
| eof			{ EOF }
| _			{ 
                          begin
                            lexerror ("Illegal Character '" ^ 
                                      (Lexing.lexeme lexbuf) ^ "'");
			    token lexbuf
			  end }
 
and str = parse	
        '"'             { ""} (* '"' *)

|	escape		{let this = Lexing.lexeme lexbuf in 
                         let rest = str lexbuf in
                         let cur = scan_escape (String.sub this 1 1) in
                         cur ^ rest }

|	_		{let cur = Lexing.lexeme lexbuf in 
                         cur ^  (str lexbuf)} 


and endline = parse 
        '\n' 		{ E.newline (); token lexbuf}
|       eof             { EOF }
|	_		{ endline lexbuf}



{

 

let parser = 
  { S.language    = "cvc";
    S.initParser        = (fun _ _ -> Cvc.initParser ());
    S.nextFewInputs = (fun lexbuf -> 
                        Parsing.clear_parser ();
                        Cvc.parseNextInput (fun _ -> Cvcparse.input 
                                                         token lexbuf));
  } 

}
