module E = Errormsg
module I = Input
module L = Logic
module S = Satproc
module P = Proof
module U = Util

open Pretty
let debug = ref false

let timingFlag = ref false

let engineIdent = "EG"                  (* This is the identifier of the
                                         * engine itself *)

let printInfix = ref true

(* Define new printers for expressions and predicates *)
let d_exp () (e: L.exp) = 
  L.printExp 
    (if !printInfix then L.infixLogicPrinter else L.defaultLogicPrinter)
    e

let d_pred () (p: L.pred) = 
  L.printPred 
    (if !printInfix then L.infixLogicPrinter else L.defaultLogicPrinter)
    p

type proofgen =
    {   start: unit -> unit;
        startVC: string -> unit;
        assume: Util.symbol -> Logic.pred -> unit;
        unassume: unit -> unit;
        pick: string -> unit;
        unpick: unit -> unit;
        conjunct: Logic.pred -> Proof.proof -> unit;
        choose: thechoice:int -> nrchoices:int -> unit;  
        finish: unit -> unit;
        finishVC: unit -> unit
      }

exception Failure

class type uiType = object
  method name: string
  method options: Options.optionDescr list
  method enterBlock: (Util.symbol * Logic.pred) list -> 
                     (string * string) list -> unit
  method exitBlock: unit

  (** Called when a goal is initiated *)
  method start_goal: Logic.pred -> unit
      

  (** Called when a goal is successfully proved *)
  method goal_success: Logic.pred -> Proof.proof -> unit

  (** Called when a goal fails *)
  method goal_failure: Logic.pred -> unit

  (** Called when an assumption is encountered *)
  method assume: Util.symbol -> L.pred -> unit

  (** Called when an assumption is retracted *)
  method unassume: unit

  (** Called to start the UI *)  
  method run: unit
end

class defaultUI : uiType = object
  method name = ""
  method options = []
  method enterBlock _ _ = ()
  method exitBlock = ()
  method start_goal _ = ()
  method goal_success _ _ = ()
  method goal_failure _ = ()
  method assume _ _ = ()
  method unassume = ()
  method run = ()
end

let theUI = ref (new defaultUI)

let continue_on_error = ref false

let continue_without_asking = ref false

(*************)
                      (* The engine will make assertions and each assertion 
                       * will be assigned a unique id. The proofs of these 
                       * assertions are kept in a hash table. We keep an undo 
                       * stack where each element specifies *)


  (* The engine state *)
type state = 
    { mutable undo:     undoElem list;
      mutable hyps:     (U.symbol * L.pred) list; (* Current hypotheses (each 
                                                * with its name *)

                                               (* These are the borken 
                                                  * apart hypotheses *)
      mutable assumes:  (int * (* Message id *)
                         L.pred * (* The assumption *)
                         bool * (* Was it a goal *) 
                         P.proof (* The proof of the assumptions *)
                           ) list;

      mutable countAssume: (int ref) list; (* For each hypothesis, how many 
                                            * independent assumes were sent 
                                            * to the proof gen *)

      mutable vars:     L.var list; (* Current variables *)
      mutable msgid:    int;
              satprocs: S.entries list;
              proofgen: proofgen;
    } 

                                        (* The engine maintains an undoStack *)
and undoElem = 
    Snapshot  of (U.symbol *  L.pred) list (* The state of the hyps *)
               * (int * L.pred * bool * P.proof) list (* The state of the 
                             1                          * assumes *)

  | Assume of int * L.pred * bool * P.proof
                                      (* Marks an assumption along with the 
                                       * message id and its proof. *)

      
                                        (* This is the proof generator *)
let proofGen (state: state) (msg: S.message) : P.proof = 
  let ownerid = msg.S.owner in
                                        (* Check first if an Engine message *)
  if ownerid = engineIdent then
    let msgid = msg.S.id in
    let rec loop = function
        Assume (id,_,_,prf) :: _ when id = msgid -> prf
      | _ :: rest -> loop rest
      | [] -> E.s (E.bug "Engine cannot find proof for message %a" 
                     S.d_message msg)
    in
    loop state.undo
  else 
    let owner = 
      try List.find (fun sp -> sp.S.ident = ownerid) state.satprocs 
      with Not_found -> E.s (E.bug "cannot find satproc %s" ownerid)
    in
    owner.S.proveIt msg

let start satprocs proofgen = 
  let state = 
    { undo     = [];
      hyps     = [];
      assumes  = [];
      countAssume = [];
      vars     = [];
      msgid    = 0;
      satprocs = satprocs;
      proofgen = proofgen;
    } in
  S.proveIt := (proofGen state);
  state.proofgen.start ();
  state


(* in debug mode, print a log string *)
let debugLog (s: string) : unit =
begin
  if !debug then
    (ignore (E.log "EG: %s\n" s));
end

let debugLogp (s: string) (p: L.pred) : unit =
begin
  if !debug then
    (ignore (E.log "EG: %s: %a\n" s L.d_pred p));
end



let finish (state: state) =
  state.proofgen.finish ()


let snapshot state = 
  if !debug then
    ignore (E.log "EG: snapshot\n");
  List.iter (fun sp -> sp.S.snapshot ()) state.satprocs;
  state.undo <- Snapshot (state.hyps, state.assumes) :: state.undo

                                        (* The global undo function *)
let undo state = 
  if !debug then
    ignore (E.log "EG: undo\n");
  List.iter (fun sp -> sp.S.undo ()) state.satprocs;
                                        (* Now pop the undoStack *)
  let rec loop = function
      Snapshot (hyps, assumes) :: rest -> 
        state.hyps <- hyps;
        state.assumes <- assumes;
        state.undo <- rest

    | Assume(_, _, _, _) :: rest -> loop rest
    | [] -> E.s (E.bug "Engine tries to pop an empty stack")
  in
  loop state.undo


(* sm: this was originally a nested function of 'broadCastAssume',
 * but I want to call it directly *)
let rec assumeMessages (state: state) (msgs: S.message list) : unit =
begin
                                      (* Send the messages to all
                                       * satprocs. Collect everything they
                                       * return *)
  let acc =
    List.fold_left
      (fun acc sp -> List.fold_left sp.S.assume acc msgs)
      []
      state.satprocs
  in
  match acc with			(* Do this until no more messages *)
    [] -> ()
  | _ -> assumeMessages state (List.rev acc)
end

                                        (* The global assume function. Might
                                         * raise S.Contradiction. Might also
                                         * raise Proved if the assumption was
                                         * goal  *)
let broadCastAssume (state: state)
                    (pred: L.pred)
                    (isgoal: bool)
                    (aboutToFail: bool)
                    (prf: P.proof) =
                                        (* First must internalize the
                                         * predicate *)
                                        (* Now create a message *)
  let predmsg =
    state.msgid <- 1 + state.msgid;
    { S.body    = pred;
      S.owner   = engineIdent;          (* Owned by the engine *)
      S.id      = state.msgid;          (* Engine does not use the id's *)
      S.goal    = isgoal;
      S.aboutToFail = aboutToFail;
    }
  in
  if !debug then
    ignore (E.log "EG: %s %a (prf=%a)\n"
              (if isgoal then "prove" else "broadcast assume")
              L.d_pred pred P.d_prf prf);

  state.undo <- (Assume(state.msgid, pred, isgoal, prf))
                :: state.undo ;

  assumeMessages state [predmsg]


(* hook for matchrule *)
let tryMatchRules : (state -> unit) option ref = ref None

let announceProof (state: state)
                  (pred: L.pred)
                  (prf: P.proof) =
  (* Here we ought to check the proof *)
  if !P.checkProofs then begin
    match prf with 
      P.G("invalid", []) -> ()
    | _ -> 
        if !E.verboseFlag then
          ignore (E.log "Checking proof of %a\n" L.d_pred pred);
        P.check state.vars state.hyps prf pred
  end;

  state.proofgen.conjunct pred prf

(** Store here the last comment that you have encountered *)
let lastComment: string ref = ref ""

(* counter for alpha conversion for quantified goals *)
let alphaCount: int ref = ref 0

                                        (* Might raise Failure *)
let rec proveConjunct
             (state: state)
             (goal: L.pred) : unit =
  let printFailMessage () =
    let rec loop = function
        [] -> ()
      | Assume(_,p,true,_) :: [] -> ()  (* This must be the failed goal *)
      | Assume(_,p,false,_) :: rest ->
          ignore (E.log "  %a\n" L.d_pred p);
          loop rest;
      |  _ :: rest -> loop rest
    in
    E.hadErrors := true;
    ignore (E.log "\n\n*****\n%sProof generation failure with assumptions\n" 
              (if !lastComment = "" then "" else "At " ^ !lastComment ^ "\n"));
    loop (List.rev state.undo);
    !theUI#goal_failure goal;
    ignore (E.log "*****@!Goal was %a@!@!" L.d_pred goal);
    (* See if we must continue *)
    let doFail () =
      (* Invoke the GUI *)
      !theUI#run;

      (* Now turn off debugging so we don't see anything after this *)
      List.iter (fun sp -> sp.S.debug := false) state.satprocs;
      raise Failure
    in
    if !continue_on_error then begin
      let answer =
        if not !continue_without_asking then begin
          print_string "*** More (y/Y) (RET to stop): ";
          read_line ()
        end else 
          "y"
      in

      if answer = "" then 
        doFail ()
      else begin
        let a = String.sub answer 0 1 in 
        (if a = "Y" then 
          continue_without_asking := true
        else if a = "y" then 
          () 
        else
          doFail ());

        ignore (E.log "\n*** Continuing after error ***\n");
        P.G("invalid", [])
      end
    end else
      doFail ()
  in
  if !debug then
    ignore (E.log "EG: conjunct @[%a@]@!" L.d_pred goal);
  match goal with 
    L.Bool false -> ignore (printFailMessage ())
  | L.Bool true -> 
      state.proofgen.conjunct goal P.Truei

  | (L.PNez _ | L.PFun _) -> begin
      snapshot state;
      let assumvar = U.registerSymbolName (L.newVar "C") in
      !theUI#start_goal goal;
      let prf = begin
        try 
          (* Can raise Contradiction or Proved *)
          broadCastAssume state goal true false (P.Hyp assumvar);
          (* We get here when we are about to fail. It is time to try 
           * harder *)
          
          
          (** Ask everybody to try harder *)
          broadCastAssume state goal true true (P.Hyp assumvar);
          printFailMessage ()
        with S.Contradiction pfalse -> P.mkContraProof assumvar goal pfalse
        | S.Proved pgoal -> pgoal
      end
      in
      !theUI#goal_success goal prf;
      announceProof state goal prf;
      (debugLogp "proved" goal);
      undo state
  end
  | L.And [goal1; goal2] -> 
      proveConjunct state goal1;
      proveConjunct state goal2

  | L.Imply(p, q) -> begin
      (* sm: prove an implication *)
      (debugLogp "impl: proving an implication" goal);

      (* --- BEGIN: copied from 'assume' --- *)
      (* Create a new assumption variable *)
      let assumprf:U.symbol = U.registerSymbolName (L.newVar "H") in
      (* Add to the hypotheses *)
      state.hyps <- (assumprf, p) :: state.hyps;
      (* --- END: copied from 'assume' --- *)

      (* let proof generation facility know that a hypothetical 
       * proof of 'p' is now available for use *)
      (state.proofgen.assume assumprf p);

      (try (
        (* assert the premise *)
        (debugLogp "impl: asserting premise" p);
        (broadCastAssume state p false false (P.Hyp assumprf));

        (* prove the consequence *)
        (debugLogp "impl: proving consequence" q);
        (proveConjunct state q);
      )
      with S.Contradiction pfalse -> (
        (* premise leads to contradiction; implication is vacuously true *)
        (debugLog "impl: contradictory premise");

        (* I think it's sufficient to simply announce the proof of false
         * as a proof of 'q', and let the 'proofgen' facility take care of
         * making the 'Impi' proof, given that I told it about 'p' already. *)
        (announceProof state q pfalse);
      ));

      (* retract hypothetical proof of 'p' *)
      (debugLog "impl: retracting premise");
      (state.proofgen.unassume ());
    end

  | L.All(var, oldPred) -> begin
      (* sm: prove universally quantified fact *)
      (debugLogp "impl: proving a universal" goal);

      (* Rename the bound variable to avoid name capture.  Use the '#'
       * character to ensure it can't collide with a user name. *)
      let newName:string = (Printf.sprintf "%s#all%d" var !alphaCount) in
      (incr alphaCount);
      let newPred:L.pred = (L.renameVariable oldPred var newName) in

      (* wrap the to-be-created proof in an 'Alli' *)
      (state.proofgen.pick newName);

      (* prove the now-renamed predicate *)
      (proveConjunct state newPred);

      (* finish the 'alli' *)
      (state.proofgen.unpick ());
    end

  | _ -> E.s (E.unimp "EG:Proving non-atoms")
      
    
   (* End the implication *)
and unassume (state: state) = 
  if !debug then
    ignore (E.log "EG: unassume@!");
  (* See how many times we must retract *)
  let countAssume = 
    match state.countAssume with 
      cnt :: rest -> state.countAssume <- rest; !cnt
    | _ -> E.s (E.bug "Not enough countAssume")
  in
  for i = 1 to countAssume do 
    state.proofgen.unassume ();
    !theUI#unassume;
  done;
  undo state


and assume (state: state) 
           (what: L.pred) 
           (nextElement: unit -> I.input) : unit = 

  snapshot state; (* One snapshot / assume *)
  (* Create a new assumption variable *)
  let assumprf = U.registerSymbolName (L.newVar "H") in
  (* Add to the hypotheses *)
  state.hyps <- (assumprf, what) :: state.hyps;
  let countAssume = ref 0 in (* We'll count separate assumes here *)
  state.countAssume <- countAssume :: state.countAssume;

  let rec mkAssumption (what: L.pred) (prf: P.proof) = 
    match what with 
      L.Bool true -> ()
    | L.And [p1; p2] -> 
        mkAssumption p1 (P.G("andel", [P.P prf]));
        mkAssumption p2 (P.G("ander", [P.P prf]))
        

    | p -> begin
        if !debug then
          ignore (E.log "EG: assume: @[%a@]@!" L.d_pred p);
      
        (** Communicate this to the oracle, to create a new axiom. We need a 
        * name  *)
        let rec mkName (prf: P.proof) : U.symbol = 
          match prf with
            P.Hyp v -> v
          | P.G(what, [P.P prf]) -> 
              U.registerSymbolName (U.symbolName (mkName prf) ^ what)
          | _ -> E.s (E.unimp "Eng.assume.mkName")
        in
        incr countAssume;
        let name = mkName prf in
        !theUI#assume name what;
        state.proofgen.assume name what;

        broadCastAssume state p false false prf
    end
  in


  (* Make the assumption, but watch for a contradiction *)
  try
    mkAssumption what (P.Hyp assumprf)
  with S.Contradiction prf_false -> begin
    if !E.verboseFlag then 
      ignore (E.log "Found contradictory assumption\n");
    (* Now prove everything until we get a matching retract *)
    let rec loopFalse (depth: int) : unit = 
      let elem = nextElement () in
(*
      if !E.hadErrors then 
        E.s (E.error "Parsing error");
*)
      match elem with
      | I.Assume what -> 
          (* We cannot ignore additional assumptions. The "impi" are 
           * expected! *)
          let assumeprf: U.symbol = U.registerSymbolName (L.newVar "H") in
          state.proofgen.assume assumeprf what;
          loopFalse (depth + 1)
            
      | I.UnAssume -> 
          if depth = 0 then begin
            (* We have retracted everything *)
            if !E.verboseFlag then 
              ignore (E.log "Finish the contradictory assumptions\n");
            unassume state
          end else begin
            state.proofgen.unassume ();
            loopFalse (depth - 1)
          end
              
      | I.EndGoal -> 
          ignore (E.warn 
                    "Found end of proof while retracting contradictory assumptions");
          ()

      | I.Prove p ->
          
          announceProof state p 
            (P.G("falsee", [P.F p; P.P prf_false]));
          
          loopFalse depth
            
      | I.Comment s -> begin 
          if !E.verboseFlag then 
            ignore (E.log "At %s@!" s); 
          lastComment := s;
          loopFalse depth
      end
      | _ -> loopFalse depth
    in
    loopFalse 0;
  end (* try *)
    

let choose (state: state) ~(thechoice: int) ~(nrchoices: int) = 
  if !debug then 
    ignore (E.log "EG: choose %d/%d\n" thechoice nrchoices);
  state.proofgen.choose thechoice nrchoices

    

let proveGoal 
    (name: string) 
    (state: state) 
    (nextElement: unit -> I.input) : unit = 

  let doOneElement (e: I.input) : unit = 
    match e with
    | I.Comment s ->
        lastComment := s;
        if !E.verboseFlag then 
          ignore (E.log "At %s@!" s)

    | I.EndGoal -> ()
        
    | I.Choose(i,nr) -> choose state i nr
          
    | I.Prove p -> proveConjunct state p

    | I.Assume ass -> assume state ass nextElement
          
    | I.UnAssume -> unassume state

    | I.Pick v -> 
        state.vars <- v :: state.vars;
        state.proofgen.pick v

    | I.Block args -> !theUI#enterBlock state.hyps args

    | I.UnBlock -> !theUI#exitBlock

    | I.UnPick -> begin
        state.proofgen.unpick ();
        match state.vars with
          _ :: rest -> state.vars <- rest
        | _ -> E.s (E.error "found UnPick but nothing was picked")
              
    end
    | I.Eof -> ()
    | _ -> E.s (E.bug "Unexpected input (%a) in goal" I.di e)
        

  in


  let rec doElements () =
    (* ignore (E.log "Looking for an element in goal %s\n" name); *)
    let elem = nextElement () in
(*
    if !E.hadErrors then 
      raise E.Error;
*)
    (* ignore (E.log "EG: Doing an element %a\n" I.delem elem); *)
    doOneElement elem;
    match elem with
      I.EndGoal -> () (* We are done *)
    | I.Eof -> ()
    | _ -> doElements () (* Keep going *)
  in

  if !debug then
    ignore (E.log "EG: start %s@!" name);
  L.resetVars ();
  state.proofgen.startVC name;

  doElements ();

  if !debug then
    ignore (E.log "EG: finished@!");

  state.proofgen.finishVC ()



