open Pretty
module E = Errormsg
module H = Hashtbl
module U = Util
module IH = Inthash 

open Logic

let checkProofs = ref false

let printMagicProofWarning = ref true

let printer = ref defaultLogicPrinter

let d_exp () e = printExp !printer e
let d_pred () p = printPred !printer p

(* In order to maintain a compact representation of proofs we represent a 
 * proof node as the name of the axiom along with a list of arguments. An 
 * exception is made for the hypothetical and parametric proof nodes. *)
type parg = 
    I of int                            (* An integer argument *)
  | S of string                         (* A string argument *)
  | E of exp                          (* An expression argument *)
  | P of proof                          (* A subproof *)
  | F of pred                         (* A predicate argument *)

and proof = 
    Hyp of  U.symbol                    (* A hypothesis, introduced with 
                                         * Impi, or Contra *) 
  | Truei

  | Falsee of proof 
    (** False elimination, if we have a proof of false *)

  | Andi of proof list
    (** Proves And[p1; p2; ... ] *)

  | Ande of int * proof
	(** Ande (i,pf) is a proof of p when
	   pf is a proof of And ps, where List.nth ps i = p *)

  | Alli of  var * proof
  | Alle of  exp * proof

  | Existe of string * proof

  | Impi of  U.symbol * pred * proof (* bind var to a hypothetical 
                                            * proof of the pred. The last 
                                            * argument is a proof of the 
                                            * right hand side, possibly 
                                            * refering to var *)
  | Impe of proof * proof

                                        (* Bind var to a proof of the 
                                         * negation of pred. The second 
                                         * argument is a proof of false *)
  | Contra of U.symbol * pred * proof
    
  | G of string * parg list             (* This is a generic proof 
                                         * constructor *)

  | Refl of exp 
  | Symm of proof
    (** Symm p proves e1=e2 if p:e2=e1 *)

  | Trans of exp * proof * proof
    (** Trans(et, p1t, p2t) proves e1=e2 if p1t:e1=et, pt2:et=e2 *)

  | Cong of exp * exp * proof list
  (** Cong \[pf\] proves Unop e1 = Unop e2 if pf proves e1 = e2;
      Cong \[pf1; pf2\] proves Binop (a1,b1) = Binop (a2,b2) if
        pf1 proves a1=a2 and pf2 proves b1=b2;
      Cong pfs proves Fun(f,as) = Fun(f,bs) if
        pfs.(n) proves as.(n) = bs.(n) 
     
      Note that each subproof of equality is really a proof 
      of (Var 0)=(Var 1) in the context in which (Var 0) and (Var 1)
      are replaced with the two things in their appropriate contexts.
   *)

  | WithLemma of Util.symbol   (* The name of the lemma *)
                * exp list (* The expressions to instantiate the universally 
                            * quantified variables *)
                * proof list (* The proofs of the hypotheses *)


  | UnrollPDef of pred * proof (* Prove a predicate of the form 
                                * "PFun(n,args)" by replacing it with its 
                                * definition and providing a proof for the 
                                * resulting expansion *)

  | UnrollEDef of Logic.pred * string * proof
      (* Prove a predicate that uses an abbreviation expression constructor.
       * The given predicate should be the goal without the substitutions made.
       *)
  | RollEDef of Logic.pred * string * proof
      (* Like UnrollEDef, but proves the substituted version of the predicate *)

  | Arith of exp * exp         (** Check that e = e' by verifying that (e - 
                                * e') mod32 = 0 using the rules of algebra. *)
  | ArithNeq of int * exp * exp   (** Check that e <> e' by verifiying that e 
                                   * - e' mod32 = i (i <> 0) using the rules 
                                   * of algebra *)
  | ArithApart of int * exp * exp (** ArithApart(r, e1, e2) means 
                                         * that e1 - e2 = r and -4 <= r <= 4.*)
  | ArithGeq of int * exp * exp  (* Check that e >= e', by verifying that e - 
                                  * e' = i and i >= 0. !!! This is unsound in 
                                  * presence of overflow. We'll have to clean 
                                  * this up. *)
  | ArithGt of int * exp * exp   (* Check that e >= e', by verifying that e - 
                                  * e' = i and i > 0. !!! This is unsound in 
                                  * presence of overflow. We'll have to clean 
                                  * this up. *)


  | SelUpd0 of Logic.exp * Logic.exp * Logic.exp 
        (* (SelUpd0 A V M) proves that (= V (sel A (upd A V M))) *)
  | SelUpd1 of Logic.exp * Logic.exp * proof
        (* (SelUpd1 V M D) proves that (= (sel A (upd A' V M)) (sel A M)) 
           provided that D proves A <> A' *)


  | Magic of string * pred (** These are proofs that are allowed to go 
                            * through the proof checker, with a warning. Pass 
                            * a string to identify this magic *)

  | MagicLemma of lemma
    (** A magic proof of a lemma *)

  | Unimplemented of string             (* We'll expand the proof language as 
                                         * needed in the future. Use 
                                         * Unimplemented until then *)


and lemma = 
    { 
      lname: U.symbol; (* The symbol corresponding to the name of the lemma *)
      lvars: string list; (* A list of variables that are universally 
                           * quantified over. These are bound in the rest of 
                           * the lemma, using UVar *)
      lhyps: Logic.pred list; (* A list of hypotheses. *)
      lconc: Logic.pred; (* A conclusion *)
    } 


(**************** Lemmas ************)

(* We keep a global registry of lemmas *)
let allLemmas: lemma IH.t = IH.create 17

let registeredLemmas () : lemma list = 
  IH.fold (fun _ l acc -> l::acc) allLemmas []

let findLemma (l: lemma) : unit = 
  try 
    if l <> IH.find allLemmas l.lname then
      E.s (E.error "Using lemma %s that is different from the registered one"
             (U.symbolName l.lname))
  with Not_found -> 
    E.s (E.error "Using unknown lemma %s\n" (U.symbolName l.lname))

(** Register a lemma, along with its proof *)
let registerLemma
    (l: lemma)
    (prf: proof)   (* This is the proof that 
                            ALL lvars. (AND lhyps) => lconc 
                    *)
    : unit = 
  if IH.mem allLemmas l.lname then 
    E.s (E.error "We have already registered lemma with name %s\n" 
           (U.symbolName l.lname));
  (* Here we should check the proof of the lemma *)
  (* ignore (E.warn "Ought to check the proof of the lemma %s!" l.lname); *)
  IH.add allLemmas l.lname l


let findRegisteredLemma (n: U.symbol) : lemma = 
  IH.find allLemmas n

(***** Pretty printing *****)
let rec d_prf () = function
    Hyp v -> text (U.symbolName v)
  | Truei -> text "truei"

  | Falsee d -> dprintf "(@[falsee %a@])" d_prf d

  | Andi pfs -> text "(andi " ++ align ++ 
	(docList ~sep:(chr ',' ++ break) (d_prf ()) () pfs) ++
	text ")" ++ unalign
  | Ande (idx,pf) -> text "(ande " ++ align ++ 
        num idx ++ break ++
	(d_prf () pf) ++ text ")" ++ unalign

  | Impi(v,p1,d2) -> 
      dprintf "(@[impi [%s : pf %a]@?%a@])" (U.symbolName v) d_pred p1 d_prf d2
  | Impe(p1,p2) -> 
      dprintf "(@[impe %a@?%a@])" d_prf p1 d_prf p2
        
  | Alli(v,d) -> 
      dprintf "(alli @[([%s : exp]@!%a)@])" v d_prf d
  | Alle(x,d) ->
      dprintf "(alle @[([%a : exp]@!%a)@])" d_exp x d_prf d

  | Existe(v,d) ->
      dprintf "(existe @[([%s : exp]@!%a)@])" v d_prf d

  | Contra(v,a,d) -> begin
      let preamble = 
        match a with
          PNez(BinOp(bop, e1, e2)) -> dprintf "%s %a %a"
              (binop_to_string bop) d_exp e1 d_exp e2
        | _ -> E.s (E.unimp "Unsupported contradiction proof")
      in
      dprintf "(con@[tra%a @!([%s : pf _]@! %a@]))" insert preamble 
        (U.symbolName v) d_prf d
  end
 
  | G(name, args) ->
      if 0 = List.length args then
        text name
      else
        dprintf "(%s @[%a@])" name (docList ~sep:break dpa) args
  | Unimplemented n -> dprintf "(unimplemented %s)" n

  | WithLemma (ln, exps, prfs) -> 
      let lvars = 
        try (findRegisteredLemma ln).lvars
        with Not_found -> 
          Util.list_init (List.length exps) (fun _ -> "?")
      in
      dprintf "Lemma(@[%s,@!%a,@!%a@])" (U.symbolName ln)
        (docList ~sep:(chr ',' ++ line) 
           (fun (n,e) -> dprintf "%s:%a" n d_exp e)) 
        (try List.combine lvars exps with Invalid_argument s ->
	  failwith ("List length mismatch while printing use of lemma " 
                    ^ (U.symbolName ln)))
        (docList ~sep:(chr ',' ++ line) (d_prf ())) prfs

  | UnrollPDef(p,prf) -> 
      dprintf "UnrollPDef(@[%a,@?%a@])"
        d_pred p d_prf prf

  | UnrollEDef(p,s,prf) -> 
      dprintf "UnrollEDef(@[%a,@?%s,@?%a@])"
        d_pred p s d_prf prf
  | RollEDef(p,s,prf) -> 
      dprintf "RollEDef(@[%a,@?%s,@?%a@])"
        d_pred p s d_prf prf

  | Arith (e1, e2) -> dprintf "(@[arith %a@? %a@?@])" d_exp e1 d_exp e2
  | ArithNeq (i, e1, e2) -> 
      dprintf "(@[arithneq@?%d@?%a@?%a@])" 
        i d_exp e1 d_exp e2
  | ArithApart (r,e1,e2) ->
      dprintf "(@[arithapart@?%d@?%a@?%a@])" 
        r d_exp e1 d_exp e2
  | ArithGeq (i,e1,e2) ->
      dprintf "(@[arithgeq@?%d@?%a@?%a@])" 
        i d_exp e1 d_exp e2
  | ArithGt (i,e1,e2) ->
      dprintf "(@[arithgt@?%d@?%a@?%a@])" 
        i d_exp e1 d_exp e2

  | SelUpd0 (addr, what, mem) -> 
      dprintf "(@[selupd0@?%a@?%a@?%a@])" 
        d_exp addr d_exp what d_exp mem
  | SelUpd1 (what, mem, prfneq) -> 
      dprintf "(@[selupd1@?%a@?%a@?%a@])" 
        d_exp what d_exp mem d_prf prfneq
      
  | Magic (s, p) -> dprintf "(magic_proof %s)" s
  | MagicLemma l -> dprintf "(magic_lemma %s)" (U.symbolName l.lname)

  | Symm d -> dprintf "(symm %a)" d_prf d
  | Refl e -> dprintf "(refl %a)" d_exp e
  | Trans (e, d1, d2) -> dprintf "(@[trans %a@?%a@?%a@])"
        d_exp e d_prf d1 d_prf d2
  | Cong (e1, e2, pfs) -> 
      dprintf "(@[cong %a@?%a@?%a@])"
        d_exp e1 d_exp e2 
        (docList ~sep:break (d_prf ())) pfs

and dpa = function
    I(i) -> num i
  | S s -> text s
  | E(e) -> d_exp () e
  | F(p) -> d_pred () p
  | P(d) -> d_prf () d



(**************** Definitions ******************)
(** Expression definitions. The name of the defined constructor (use a 
  * lowercase name), the list of formals, and the body, referring to the 
  * formals using deBruijn indices. *)
type edef = string * string list * Logic.exp 

(* Predicate definitions. The name of the defined constructor (use an 
 * uppercase initial for the name), the list of formals, and the body, 
 * referring to the formals using deBruijn indices. *)
type pdef = string * string list * Logic.pred 

let allEDefs: (string, edef) H.t = H.create 13
let allPDefs: (string, pdef) H.t = H.create 13

(** Register a predicate definition *)
let registerPDef (x : pdef) : unit = 
  let pname, _, _ = x in
  if H.mem allPDefs pname then 
    E.s (E.error "We have already registered the predicate constructor %s\n" 
           pname);
  H.add allPDefs pname x

(** Register an expression definition *)
let registerEDef (x : edef) : unit = 
  let ename, _, _ = x in
  if H.mem allEDefs ename then 
    E.s (E.error "We have already registered the expression constructor %s\n" 
           ename);
  H.add allEDefs ename x




(****** Type inference *****)
exception ProofCheck of doc * proof * (U.symbol * pred) list

let showAssumptions (hyps: (U.symbol * pred) list) = 
  List.iter
    (fun (v, a) -> 
      ignore (E.log " %s: %a\n" (U.symbolName v) 
                                d_pred a))
    (List.rev hyps)

(*** This is not sound w.r.t. modulo arithmetic *)
(** Check an equality: e1 = e2 + n  *)
let checkEquality (e: exp) : bool = 
  (* Keep a hash-table of non-arithmetic expressions, with their coefficient *)
  let coeffs : (exp, int ref) H.t = H.create 13 in 
  (* Maintain the integer coefficient also *)
  let icoeff : int ref = ref 0 in
  let rec scan (factor: int) = function
      Int n -> icoeff := !icoeff + factor * (to_int n)
    | BinOp((Plus | Add), e1, e2) -> scan factor e1; scan factor e2
    | BinOp((Minus | Sub), e1, e2) -> scan factor e1; scan (- factor) e2
    | BinOp((Times | Mult), Int n, e) -> scan (to_int n * factor) e
    | BinOp((Times | Mult), e, Int n) -> scan (to_int n * factor) e
    | BinOp(Sll, e, Int n) -> scan (factor lsl (to_int n)) e
    | e -> begin
        let c : int ref = 
          try
            H.find coeffs e
          with Not_found -> begin
            let c = ref 0 in
            H.add coeffs e c;
            c
          end
        in
        c := !c + factor
    end
  in
  scan 1 e;
  (* Check that we have simplified everything to 0 *)
  let err = ref nil in
  if !icoeff <> 0 then 
    err := dprintf "  Constant coefficient is %d\n" !icoeff;
  H.iter 
    (fun e c -> 
      if !c <> 0 then 
        err := dprintf "%a Coefficient of %a is %d\n" insert !err d_exp e !c)
    coeffs;
  if !err != nil then begin
    ignore (E.warn "Cannot verify zeroness of %a\n%a" d_exp e insert !err);
    false
  end else
    true

let makeExistVar n = Fun ("existVar", [Fun (n, [])])

let existTbl : (string, pred) Hashtbl.t = Hashtbl.create 100
(* A table saving which "new" terms were used with which existential
 * formulas, to avoid an unsound repeated instantiation. *)

let rec infer (vars: string list) 
              (hyps: (U.symbol * pred) list) 
              (d: proof)  : pred = 
  let rec i (d: proof) =
    match d with 
    | Truei -> Bool true

    (* Magic *)
    | Magic (s,p) ->
	if !printMagicProofWarning then 
          ignore (E.warn "Magic proof (%s) for %t.\n" s 
		    (fun () -> begin try
		      ignore (Str.search_forward 
				(Str.regexp "HIDEPRED") s 0);
		      text "<pred omitted>"
		    with Not_found ->
		      d_pred () p
		    end));
	p

    | WithLemma (lname, exps, prfs) -> begin
        (* Lookup the lemma *)
        let lemma =
          try findRegisteredLemma lname
          with Not_found -> E.s (E.error "Cannot find registered lemma %s"
                                   (U.symbolName lname))
        in
        (* Check that we have enough instantiations *)
        if List.length exps <> List.length lemma.lvars then 
          E.s (E.error "Lemma %s expects %d instantiations and is used here with :\n\t@[%a@]\n"
                 (U.symbolName lemma.lname) 
                 (List.length lemma.lvars)
                 (docList ~sep:(chr ',' ++break) (d_exp ())) exps);

        (* Now construct a substitution context. Each of the instantiations 
         * uses the current context *)
        let subst = newSubstFromList exps in
	let theconc : pred = 
          try unrollUVarsPredRaise subst lemma.lconc
          with HasUVar -> 
            E.s (E.bug "Not enough instantiations for conclusion of lemma %s" 
                   (U.symbolName lemma.lname))
        in

        (* Now check the proofs of the hypotheses. We could use List.iter2, 
         * but this is faster, allows us to check for different list lengths, 
         * and does not confuse the profiler. *)
        let rec checkHyps (hyp_prfs: proof list) (lhyps: pred list) = 
          match hyp_prfs, lhyps with
            [], [] -> ()
          | hyp_prf :: rest_hyp_prfs, lhyp :: rest_lhyps -> 
              let hyp' = 
                try unrollUVarsPredRaise subst lhyp 
                with HasUVar -> 
                  E.s (E.bug "Not enough instantiations for hypothesis of lemma %s" 
                         (U.symbolName lemma.lname))
              in
              let hyp'' = i hyp_prf in
              if hyp' <> hyp'' then begin
                let msg = 
                  dprintf "Hypothesis %d of lemma %s: Cannot check that we prove@!          %a@! we prove %a\n" 
                    (List.length lemma.lhyps - List.length rest_lhyps)
                    (U.symbolName lemma.lname) 
                    d_pred hyp' d_pred hyp'' 
                in 
                raise (ProofCheck (msg, hyp_prf, hyps))
              end;
              checkHyps rest_hyp_prfs rest_lhyps

          | _, _ -> 
              E.s (E.error "Lemma %s expects %d subproofs and is used here with %d:\n\t@[%a@]\n"
                     (U.symbolName lemma.lname) 
                     (List.length lemma.lhyps)
                     (List.length prfs)
                     (docList ~sep:line (d_prf ())) prfs)
        in
        checkHyps prfs lemma.lhyps;
        
        theconc
    end
	  
    | UnrollPDef (prd, prf) -> begin
        let pname, pargs, (_, pd_formals, pd_body) = 
          try
            match prd with 
	      PFun(pname, pargs) -> pname, pargs, H.find allPDefs pname
            | _ -> raise Not_found
          with Not_found -> 
            E.s (E.error "UnrollPExp used on an undefined predicate %a"
                   d_pred prd)
        in
        let subst = newSubstFromList pargs in

        if List.length pd_formals <> List.length pargs then 
          E.s (E.error "Predicate constructor %s expects %d arguments and is used here as follows:\n\t%a\n"
                 pname (List.length pd_formals)
                 d_pred prd);

        (* The predicate to be proved *)
	let prd' : pred = 
          try unrollUVarsPredRaise subst pd_body
          with HasUVar -> 
            E.s (E.bug "Not enough instantiations for conclusion of defined predicate %s" pname)
        in
        
        let prd'' = i prf in
        
        if prd' <> prd'' then begin
          let msg = dprintf "UnrollPDef %s does not prove %a\nIt proves: %a\n"
              pname d_pred prd'
	      d_pred prd'' in
          raise (ProofCheck (msg, prf, hyps))
        end;

        prd
    end
	
    | RollEDef (prd, ename, prf) -> begin
        let _, ed_formals, ed_body = 
	  try
	    H.find allEDefs ename
          with Not_found -> 
            E.s (E.error "UnrollEDef used with an undefined constructor %s"
                   ename)
        in

	(match ed_formals with
	     [] -> ()
	   | _ -> failwith "Unimplemented: Non-nullary expression shortcuts");

	let subber depth = function
	    Fun (f, []) when f = ename -> Some ed_body
	  | _ -> None in

	let prd' = substPred subber prd in

        let prd'' = i prf in
        
        if prd <> prd'' then begin
          let msg = dprintf "UnrollEDef %s does not prove %a\nIt proves: %a\n"
                 ename d_pred prd
		 d_pred prd'' in 
          raise (ProofCheck (msg, prf, hyps))
        end;

        prd'
    end
    | UnrollEDef (prd, ename, prf) -> begin
        let _, ed_formals, ed_body = 
	  try
	    H.find allEDefs ename
          with Not_found -> 
            E.s (E.error "UnrollEDef used with an undefined constructor %s"
                   ename)
        in

	(match ed_formals with
	     [] -> ()
	   | _ -> failwith "Unimplemented: Non-nullary expression shortcuts");

	let subber depth = function
	    Fun (f, []) when f = ename -> Some ed_body
	  | _ -> None in

	let prd' = substPred subber prd in

        let prd'' = i prf in
        
        if prd' <> prd'' then begin
          let msg = 
            dprintf "UnrollEDef %s does not prove %a\nIt proves: %a\n"
              ename d_pred prd'
	      d_pred prd'' in 
          raise (ProofCheck (msg, prf, hyps))
        end;

        prd
    end

    (* And *)
    | G("andi",  [P d1; P d2]) -> And [i d1; i d2]
    | G("andel", [P d]) -> begin
        match i d with
          And (p1:: _) -> p1
        | _ -> raise (ProofCheck (text "andel does not prove an And",
                                  d, hyps))
    end
    | G("ander", [P d]) -> begin
        match i d with
          And (_ :: p2 :: _) -> p2
        | _ -> raise (ProofCheck (text "ander does not prove an And",
                                  d, hyps))
    end
    | Andi ds -> 
        (* We define our own List.map to not confuse the Profiler *)
        let rec subproofs = function
            [] -> []
          | d :: ds -> i d :: subproofs ds
        in
        And (subproofs ds)

    | Ande (idx,d) -> begin
	match i d with
	  And ds -> List.nth ds idx
	| _ -> raise (ProofCheck (text "Ande does not prove an And",
                                  d, hyps))
    end

    | G("oril", [P d; F other]) -> Or [i d; other]
    | G("orir", [F other; P d]) -> Or [other; i d]

    | Alli(v,d) -> All(v, infer (v :: vars) hyps d)
    | Alle (x, pf) ->
	(match infer vars hyps pf with
	     All (v, body) ->
	       let name = Some x in
	       let subber depth = function
		   Var i ->
		     if i - depth = 0 then
		       name
		     else
		       None
		 | _ -> None in
	       let body' = substPred subber body in
	       body'
	   | _ -> raise (ProofCheck (text "Alle does not prove All",
                                     pf, hyps)))

    | Existe (name, pf) ->
	(match infer vars hyps pf with
	     Exist (v, body) ->
	       (match Hashtbl.find_all existTbl name with
		    [] ->
		      (* This is a fresh name *)
		      Hashtbl.add existTbl name body
		  | p::_ ->
		      if p <> body then begin
			let msg = dprintf "Duplicate instantation %s for Existe.\nPreviously used with: %a\nUsed now with: %a\n"
			       name d_pred p d_pred body in 
			raise (ProofCheck (msg, pf, hyps))
		      end);

	       let name = Some (makeExistVar name) in
	       let subber depth = function
		   Var i ->
		     if i - depth = 0 then
		       name
		     else
		       None
		 | _ -> None in
	       let body' = substPred subber body in
	       body'
	   | _ -> raise (ProofCheck (text "Existe does not prove Exist",
                                     pf, hyps)))
       
    | Impi(v,p1,d2) -> Imply(p1, 
                             infer vars ((v,p1) :: hyps) d2)
    | Impe(d1,d2) ->
	(match infer vars hyps d1 with
	     Imply (hyp, conc) ->
	       let prd = infer vars hyps d2 in
		 if prd <> hyp then begin
		   let msg = dprintf 
                       "Impe needs hypothesis %a\nBut given: %a\n"
			  d_pred hyp
			  d_pred prd in
		   raise (ProofCheck (msg, 
                                      d2, hyps))
		 end;
		 conc
	  | prd ->
	      let msg = dprintf "Impe given proof of non-implication: %a\n"
		     d_pred prd in 
	      raise (ProofCheck (msg, d2, hyps)))

    | Contra(v,p,d1) -> begin
        match infer vars ((v, negatePred p) :: hyps) d1 with
          Bool false -> p
        | _ -> raise (ProofCheck (text "Contra does not prove false",
                                  d, hyps))
    end
    | (Hyp v) -> begin
        try 
          let (_, p) = List.find (fun (v',_) -> v = v') hyps in
          p
        with Not_found -> raise (ProofCheck (text "Hyp not found",
                                             d, hyps))
    end
 
    (* false elimination *)
    | G("falsee", [ F p ; P pff ]) -> begin
        match i pff with
          Bool false -> p
        |   _ -> raise (ProofCheck (nil, d, hyps))
    end

    | G("doselmap", [ E m; I idx]) -> begin
        try
          let e = findMap m idx in
          eq e (Fun("selmap", [ m; integer idx]))
        with Not_found -> begin
          let msg = dprintf "doselmap: cannot find index %d\n" idx in 
          raise (ProofCheck (msg, d, hyps))
        end
    end

    | G("doupdmap", [ E m; I idx; E newval ]) -> begin
        let m' = insertMap m idx newval in
        eq m' (Fun("updmap", [ m; integer idx; newval ]))
    end
          
          (** New style arithmetic *)
    | Arith (e1, e2) ->
        if checkEquality (minus e1 e2) then 
          eq e1 e2
        else begin
          let msg = dprintf "Arith: checkEquality 0= %a - %a failed"
              d_exp e1 d_exp e2 in 
          raise (ProofCheck (msg, d, hyps))
        end

    | ArithNeq (r, e1, e2) -> 
	if checkEquality (minus (minus e1 e2) (integer r)) && r <> 0 then 
	  ne e1 e2
	else begin 
          let msg = dprintf "ArithNeq checkEquality failed" in 
	  raise (ProofCheck (msg, d, hyps))
        end

    | ArithGeq (r, e1, e2) ->
	if checkEquality (minus (minus e1 e2) (integer  r)) && r >= 0 then 
	  geq e1 e2
	else begin
          let msg = dprintf "ArithGeq: checkEquality failed" in 
	  raise (ProofCheck (msg, d, hyps))
        end

    | ArithGt (r, e1, e2) ->
        let z = minus (minus e1 e2) (integer  r) in
	if checkEquality  z && r > 0 then 
	  gt e1 e2
	else begin
          let msg = 
            dprintf "ArithGt: cannot check equality of %a\n" d_exp z in 
	  raise (ProofCheck (msg, d, hyps))
        end


    | ArithApart (r, e1, e2) -> 
        if checkEquality (minus (minus e1 e2) (integer r)) && 
           (r <= -4 || r >= 4) then
	  PFun("Apart", [e1; e2])
        else begin 
          let msg = dprintf "ArithApart" in 
          raise (ProofCheck (msg, d, hyps))
        end


    | SelUpd0 (addr, what, mem) -> 
        eq what (Fun("sel", [ addr; Fun("upd", [ addr; what; mem ])]))
    | SelUpd1(what, mem, prfneq) -> begin
        match i prfneq with 
          PNez(BinOp(SetNe, seladdr, updaddr)) -> 
            eq (Fun("sel", [seladdr; 
                            Fun("upd", [ updaddr; what; mem])]))
               (Fun("sel", [seladdr; mem ]))
        | _ -> 
            let msg = dprintf "selupd1: cannot prove disequality" in
            raise (ProofCheck (msg, d, hyps))
    end
        
	  (* arithmetic axioms *)
    | G ("eqCheck", [E e1; E e2]) ->
	if checkEquality (minus e1 e2) then 
	  eq e1 e2
	else begin 
          let msg = dprintf "eqCheck" in 
	  raise (ProofCheck (msg, d, hyps))
        end


    | G ("geqCheck", [E e1; E e2; I r]) ->
        let z = minus (plus e1 (integer  r)) e2 in
	if checkEquality z && 0 >= r then 
	  geq e1 e2
	else begin
          let msg = dprintf "geqCheck: cannot check equality of %a\n" d_exp z 
          in
	  raise (ProofCheck (msg, d, hyps))
        end


    | G("geq0", []) -> geq zero zero
          
    | G("eq0", []) -> eq zero zero 
          
    | G("geq*>=", [I n; P d1; P d2]) -> begin
        if n < 1 then begin
          let msg = dprintf "factor should be strictly positive in geq*>=" in 
          raise (ProofCheck (msg, d, hyps))
        end;
	match i d1, i d2 with
	  (PNez(BinOp(SetGe, e1,e2)), 
           PNez(BinOp(SetGe, rest, z))) when z = zero -> begin
	     geq (plus (times 
                          (integer n) 
                          (minus e1 e2))
                    rest) zero
	   end
	| _ -> raise (ProofCheck (text "geq*>= does not prove >=",
                                  d, hyps))
    end
    | G("geq*=", [I n; P d1; P d2]) -> begin
	match i d1, i d2 with
	  (PNez (BinOp(SetEq, e1,e2)), 
           PNez (BinOp(SetGe, rest, z))) when z = zero -> begin
	     geq (plus (times (integer n) 
                               (minus e1 e2))
                    rest)
               zero
	   end
	| _ -> raise (ProofCheck (text "geq*= does not prove >=",
                                  d, hyps))
    end

    | G("geq*i", [E e1; E e2; I r; I n; P d1; P d2]) -> begin
	match i d1, i d2 with
	  PNez (BinOp(SetEq, 
                      BinOp(Plus, 
                            BinOp(Times, 
                                  Int n', 
                                BinOp(Minus, e1',e2')), 
                            BinOp (Minus, Int r', o)), 
                      rest)),
          PNez (BinOp(SetGe, rest', z)) when o = one && z = zero ->
            let n' = to_int n' in
            let r' = to_int r' in
	    if n=n' && r=r' && n>=r && n>= 1 && e1=e1' && e2=e2' 
                    && rest=rest' then
	      geq e1 e2
            else begin 
              let msg = dprintf "geq*i: " in
	      raise (ProofCheck (msg, d, hyps))
            end
	| p1, p2 -> begin
            let rest = Fun("_", []) in
            let msg = 
              dprintf "d1 proves: %a\n   expected: %a\n  d2 proves: %a\n   expected: %a\n"
                d_pred p1
                d_pred
                (PNez (BinOp(SetEq, 
                             BinOp(Plus, 
                                   BinOp(Times, 
                                         integer n, 
                                         BinOp(Minus, e1,e2)), 
                                   BinOp (Minus, integer r, one)), 
                             rest)))
                d_pred p2
                d_pred 
                (PNez (BinOp(SetGe, rest, zero))) in
            raise (ProofCheck (msg, d, hyps))
        end
    end
       
    | G("eq*=", [I n; P d1; P d2]) -> begin
	match i d1, i d2 with
	  (PNez (BinOp(SetEq,e1,e2)), 
           PNez (BinOp(SetEq, rest, z))) when z = zero -> begin
	     eq (plus (times (integer n)
                         (minus e1 e2))
                   rest)
               zero
	   end
	| _ -> raise (ProofCheck (text "eq*=", d, hyps))
    end

    | G("eq*i", [E e1; E e2; I n; 
                  P d_ege_ele; 
                  P d_total;
                  P d_ege;
                  P d_ele;
                  P d_eq]) 
      -> begin
        let ege, ele = 
	  match i d_ege_ele with
            PNez (BinOp(SetEq, BinOp(Plus, ege, ele), z)) when z = zero 
            -> ege, ele
          | _ -> raise (ProofCheck (text "eq*i: 1", d, hyps))
        in
        (match i d_ege, i d_ele with 
          PNez (BinOp(SetGe, ege', z)), 
          PNez (BinOp(SetGe, ele', z')) 
          when z = zero && z' = zero && ege = ege' && ele = ele' -> ()
        | _ -> raise (ProofCheck (text "eq*i: 2", d, hyps)));
        
        let eqq = 
          match i d_eq with 
            PNez (BinOp(SetEq, eqq, z)) when z = zero -> eqq
          | _ -> raise (ProofCheck (text "eq*i: 3", d, hyps))
        in
        (match i d_total with
	  PNez (BinOp(SetEq,
                      BinOp(Times, 
                            Int n', 
                            BinOp(Minus, e1',e2')), 
                      BinOp(Plus, ege', eq'))) 
            when n = to_int n' && 
                 e1 = e1' && e2 = e2' && ege = ege' && eqq = eq' -> ()
        | _ -> raise (ProofCheck (text "eq*i: 4", d, hyps)));

        eq e1 e2
    end
       
    | G("neqGt", [P d1; P d2]) -> begin
        match i d1, i d2 with
          PNez (BinOp(SetGe, e1, e2)), 
          PNez (BinOp(SetNe, e1', e2')) when e1 = e1' && e2 = e2' -> 
            geq e1 (plus e2 one)
        | p1, p2 -> 
            let msg = dprintf "neqGt: %a@!   %a@!" d_pred p1 d_pred p2 in 
            raise (ProofCheck (msg, d, hyps))
    end   
    | G("eqi", [P d1; P d2]) -> begin
	match (i d1, i d2) with
	  (PNez (BinOp(SetGe, e1,e2)), 
           PNez (BinOp(SetGe, e_2, e_1))) ->
	     if (e1 = e_1) && (e2 = e_2) then begin
               eq e1 e2
	    end else
	      raise (ProofCheck (text "eqi: 1", d, hyps))

	| _ -> raise (ProofCheck (text "eqi: 2", d, hyps))
    end

    | G("eqgeq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetEq, e1,e2)) -> geq e1 e2
	| _ -> raise (ProofCheck (text "eqgeq", d, hyps))
    end

    | G("eqgeqr", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetEq, e1,e2)) -> geq e2 e1
	| _ -> raise (ProofCheck (text "eqgeqr", d, hyps))
    end
    | G("ltgeq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetLt, e1,e2)) -> geq e2 (plus e1 one)
	| _ -> raise (ProofCheck (text "ltgeq", d, hyps))
    end
    | G("geqlt", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGe, e2, BinOp(Plus, e1, o))) when o = one ->
            lt e1 e2
	| _ -> raise (ProofCheck (text "geqlt", d, hyps))
    end
    | G("leqgeq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetLe, e1,e2)) -> geq e2 e1
	| _ -> raise (ProofCheck (text "leqgeq", d, hyps))
    end
    | G("geqleq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGe,e1,e2)) -> leq e2 e1
	| _ -> raise (ProofCheck (text "geqleq", d, hyps))
    end
    | G("gtlt", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGt,e1,e2)) -> lt e2 e1
	| _ -> raise (ProofCheck (text "gtlt", d, hyps))
    end

    | G("gtgeq", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGt, e1,e2)) -> geq e1 (plus e2 one)
	| _ -> raise (ProofCheck (text "gtgeq", d, hyps))
    end

    | G("geqgt", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGe, e1, BinOp(Plus, e2, o))) when o = one ->
            gt e1 e2
	| _ -> raise (ProofCheck (text "geqgt", d, hyps))
    end

    | G("eqeq0", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetEq, e1, e2)) -> eq (minus e1 e2) zero
	| _ -> raise (ProofCheck (text "eqeq0", d, hyps))
    end

    | G("eqgeq0", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetEq, e1, e2)) -> geq (minus e1 e2) zero
	| _ -> raise (ProofCheck (text "eqgeq0", d, hyps))
    end

    | G("geqgeq0", [P d1]) -> begin
	match i d1 with
	  PNez (BinOp(SetGe, e1, e2)) -> geq (minus e1 e2) zero
	| _ -> raise (ProofCheck (text "geqgeq0", d, hyps))
    end

    | G("geqLt_FI", [E e1; E e2; P d1; P d2]) -> begin
	match (i d1, i d2) with
	  (PNez (BinOp(SetGe, e1',e2')),
           PNez (BinOp(SetGe, e2'', 
                       BinOp(Plus, e1'', o)))) when o = one ->
	     if (e1 = e1') && (e1 = e1'') && 
                (e2 = e2') && (e2 = e2'') then
	       Bool false
	     else 
	       raise (ProofCheck (text "geqLt_FI: 0", d, hyps))

	| _ -> raise (ProofCheck (text "geqLt_FI: 1", d, hyps))
    end


	    (* Congruence begins *)
    | G ("not_FI",[P p1; P p2]) -> begin
	match ((i p1),(i p2)) with
	  (PNez (BinOp(SetEq, e1,e2)),
           PNez (BinOp(SetNe, e3,e4))) ->
	     if (e1 = e3) & (e2 = e4) then Bool false
	     else begin
               let msg = dprintf "not_FEq not right" in 
               raise (ProofCheck (msg, d, hyps))
             end
	| (PNez (BinOp(SetGe, e1,e2)),
           PNez (BinOp(SetLt, e3,e4))) ->
	     if (e1 = e3) & (e2 = e4) then Bool false
	     else begin
               let msg = dprintf "not_FGeq/Lt not right" in 
               raise (ProofCheck (msg, d, hyps))
             end
        | _ -> 
            let msg = dprintf "Proof: not_FI: ???\n" in 
            raise (ProofCheck (msg, d, hyps))
    end
	      

    | G ("eqNeq_FI",[E e1; E e2; P p1; P p2]) -> begin
	match i p1, i p2 with
	  (PNez (BinOp(SetEq, e1', e2')),
           PNez (BinOp(SetNe, e1'', e2''))) ->
	     if e1 = e1' && e1 = e1'' && 
                e2 = e2' && e2 = e2'' then 
               Bool false
	     else begin 
               let msg = dprintf "eqNeq_FI:Eq  Proof.infer says %a is not a valid proof" 
                          d_prf (G ("eqNeq_FI",[P p1;P p2])) in 
		
	        raise (ProofCheck (msg, d, hyps))
             end

	| _ -> 	
            let msg = dprintf "eqNeq_FI: ??? invalid use of axiom! in %a" 
                d_prf (G ("eqNeq_FI",[P p1;P p2])) in 
            raise (ProofCheck (msg, d, hyps))
    end
    | G ("eqid",[E e]) -> eq e e
    | Refl e -> eq e e

    | G ("eqsym",[P p]) -> begin
	match i p with
	  PNez (BinOp(SetEq, e1,e2)) -> eq e2 e1
	| _ -> 
            let msg = 
              dprintf "eqsym: invalid use of axiom! in %a" 
                d_prf (G ("eqsym",[P p])) in 
            raise (ProofCheck (msg, d, hyps))
    end

    | G ("eqtr", [E e2; P p1; P p2]) -> begin
	match i p1, i p2 with
	  (PNez (BinOp(SetEq, e1, e2')),
           PNez (BinOp(SetEq, e2'', e3))) ->
	     if e2 = e2'' && e2 = e2' then 
	       eq e1 e3
	     else 
               (ignore (E.warn "eqtr: invalid use of axiom! in %a@!e2=e2''=%b@!e2=e2'=%b@!e2=%a@!e2''=%a@!e2'=%a@!" 
                          d_prf d (e2=e2'') (e2=e2') 
                          d_exp e2 d_exp e2'' d_exp e2');
                showAssumptions hyps;
                raise (ProofCheck (nil, d, hyps)))

	| _ -> 
            let msg= dprintf "eqtr: invalid use of axiom! in %a" 
                d_prf (G ("eqtr",[P p1;P p2])) in
            raise (ProofCheck (msg, d, hyps))
    end

    | G("neqsym", [ P p1] ) -> begin
        match i p1 with 
          PNez(BinOp(SetNe, e1, e2)) -> ne e2 e1
        | _ -> 
            let msg = dprintf "premise of neqsym is not a disequality" in 
            raise (ProofCheck (msg, p1, hyps))
    end
    | G("neqtr", [ P peqa; P pneq; P peqb ]) -> begin
        let a, a' = match i peqa with
          PNez (BinOp(SetEq, a, a')) -> a, a'
        | _ -> 
            let msg = dprintf "1st premise of neqtr is not an equality" in 
            raise (ProofCheck (msg, peqa, hyps))
        in
        let b', b = match i peqb with
          PNez (BinOp(SetEq, b', b)) -> b', b
        | _ -> 
            let msg = dprintf "3rd premise of neqtr is not an equality" in 
            raise (ProofCheck (msg, peqb, hyps)) 
        in
        match i pneq with 
          PNez (BinOp(SetNe, a'', b'')) when a' = a'' && b' = b'' ->
            ne a b
        | _ -> 
            let msg = dprintf "2nd premise of neqtr does not prove %a <> %a" 
                d_exp a' d_exp b'
            in 
            raise (ProofCheck (msg, pneq, hyps)) 
    end
        
        
    | G ("congr", (S name) :: eqargs) -> 
        (* find the arguments *)
        let rec loop (eqargs: parg list) : (exp list * exp list) = 
            match eqargs with 
               [] -> [], []
             | P eqprf :: rest -> begin
                 let a1, a2 = loop rest in
                 match i eqprf with 
                    PNez (BinOp(SetEq, e1, e2)) -> e1 :: a1, e2 :: a2
                  | _ -> 
                      let msg = dprintf "Subproof of \"congr\" is not =" in 
                      raise (ProofCheck (msg, d, hyps))
             end
             | _ -> 
                 let msg = dprintf "Subproof of \"congr\" is not a proof" in 
                 raise (ProofCheck (msg, d, hyps))
         in
         let a1, a2 = loop eqargs in 
         (* We assume for now that the name corresponds to a function *)
         eq (Fun(name, a1)) (Fun(name, a2))
 

   | d -> E.s (E.unimp "Proof.infer:@!%a@!" d_prf d)
  in
    i d

let check (vars: string list) 
          (hyps: (U.symbol * pred) list) 
          (prf: proof) 
          (goal: pred) = 
  try
    let ipred = infer vars hyps prf in
    if ipred <> goal then begin
      let msg = dprintf "Proof proves a different predicate from the goal@!Got :%a@!Need:%a" d_pred ipred d_pred goal in 
      raise (ProofCheck (msg, prf, hyps))
    end
        
  with ProofCheck (msg, d, hyps) -> 
    ignore (E.warn "Invalid proof:@!%a@!of@!%a" 
              d_prf d d_pred goal);
    showAssumptions hyps;
    E.s (E.error "Proof checking failed:\n  %a\n" insert msg)

  | E.Error -> 
    ignore (E.warn "Error checking proof:@!%a" d_prf prf);
    showAssumptions hyps;
    E.s (E.error "Proof checking failed")
      


 (** Make a proof by contradiction *)
let mkContraProof (h: U.symbol)
                  (a: pred)
                  (pffalse: proof) (* [H:pf (not A)] pf false *)
    : proof  (* pf A *) =
(*
  ignore (E.log "mkContraProof %s:%a is false: %a\n"
            (U.symbolName h)
            d_pred a
            d_prf pffalse);
*)
  (* A few obvious optimizations *)
  match a, pffalse with 
    PNez(BinOp(SetEq, e1, e2)), 
    G("eqNeq_FI", [E _; E _; P pfeq; P (Hyp h')]) 
      when h = h' -> pfeq
(*
  | Geq(e1, e2), 
        G("geqLt_FI", [ E _; E _; 
                        P (G("ltgeq", [P (Hyp h')]));
                        P (G("geq*i", [ E (Plus(e1, Int 1)); 
                                        E (Plus(e2, Int 1)); 
                                        r; n; 
                                        P (G("eqCheck", 
                                             [ Ea5; a6 
                                                 I diff; I fact; 
                                               d1; d2
                                                 e1_ge_e2; 
                                 ])
          when h = h' -> e1_ge_e2
*)
  | _ -> Contra (h, a, pffalse)


(*
					(* Get an assumption *)
    fun getAss (v, (v', p') :: rest) = if v = v' then p' else getAss(v, rest)

					(* Compare a list of pairs *)
    fun allEq el = List.all E.Equiv el

					(* Emit an Arith proof *)
    fun emitArith ass (prf as (Arith arg)) =
      let
	fun doit _ = prf
      in
	doit arg
      end
      | emitArith ass prf = prf
					(* Emit a contradiction proof *)
    fun emit (v : V.T, egoal : E.T, prf : T, ass : (V.T * P.T) list) : T =
      let
	fun mkContra (prf : T) : T = Contra (v, P.PNez egoal, prf)
	val emitArith = emitArith ass

	  
	fun doEqNeq (eqprf as Ass u, neqprf as Ass u') : Proof.T =
	  if u = v then
	    neqprf
	  else if u' = v then
	        eqprf
	       else
		 mkContra (EqNeq_Falsei(eqprf, neqprf))
	  | doEqNeq (eqprf as Ass u, neqprf : T) : Proof.T =
	    if u = v then neqprf else mkContra (EqNeq_Falsei(eqprf, neqprf))
	  | doEqNeq (eqprf , neqprf as Ass u) : Proof.T =
	    if u = v then eqprf else mk                                    ])
Contra (EqNeq_Falsei(eqprf, neqprf))
	  | doEqNeq (eqprf, neqprf) = mkContra (EqNeq_Falsei(eqprf, neqprf))



	fun doGeqLt (d1 as Ass u, d2) = 
	  if u = v then
	    GeqLt (emitArith d2)
	  else
	    mkContra (GeqLt_Falsei(d1, d2))
	    
	  | doGeqLt (d1 as LeqGeq (Ass u), d2) = 
	    if u = v then
	      GeqGt (emitArith d2)
	    else
	      mkContra (GeqLt_Falsei(emitArith d1, emitArith d2))
	      
	  |  doGeqLt (d1 as GtGeq (Ass u),
		      d2 as Arith(ArithGEQ(rest,
					   E.Plus(e1, E.Int 1),
					   E.Plus(e2, E.Int 1)),
				  fact, eql,geql,leql)) =
	     if u = v then
	       GeqLeq (emitArith (Arith(ArithGEQ(rest, e1, e2),
					fact, eql, geql, leql)))
	     else
	       mkContra (GeqLt_Falsei(emitArith d1, emitArith d2))
	       
	  | doGeqLt (d1 as LtGeq (Ass u),
		     d2 as Arith (ArithGEQ(rest,
					   E.Plus(e1, E.Int 1),
					   E.Plus(e2, E.Int 1)),
				  fact, eql,geql, leql)) =
	    if u = v then
	      emitArith (Arith(ArithGEQ(rest, e1, e2), fact, eql, geql, leql))
	    else
	      mkContra (GeqLt_Falsei(emitArith d1, emitArith d2))

	  | doGeqLt (d1, d2) = mkContra (GeqLt_Falsei(emitArith d1,
						      emitArith d2))

	fun ctr (EqNeq_Falsei(eqprf, neqprf)) = 
	    doEqNeq(emitArith eqprf, emitArith neqprf)

	  | ctr (GeqLt_Falsei(geqprf, leqprf)) = doGeqLt(emitArith geqprf,
							 emitArith leqprf)
	  | ctr prf =
	    let
	      fun findInnerMost (Not_Falsei (prf, Proof.Ass v')) =
		(if v' = v then		(* Found one *)
		   case findInnerMost prf of
		     [] => [prf]
		   | l => l
		 else
		   findInnerMost prf)
		| findInnerMost d = []
	    in
	      case findInnerMost prf of
		[] => mkContra (emitArith prf)
	      | h :: _ => h
	    end
      in
	if contraOpt then
	  ctr prf
	else
	  mkContra prf
      end
  end
*)

(* We want to allow the registration of lemmas, predicate definitions and 
 * expressions definitions at the top level. We take a snapshot when we call 
 * init the first time. *)
let initThunk: (unit -> unit) option ref = ref None

let snapshot () : unit -> unit = 
  U.runThunks [ U.restoreIntHash allLemmas;
                U.restoreHash allPDefs;
                U.restoreHash allEDefs ]

let init () = 
  match !initThunk with 
    None -> initThunk := Some (snapshot ())
  | Some t -> t ()

