module E = Errormsg
module I = Input
module L = Logic

open Pretty
  
(** This defines the signature that a satisfiability procedure must implement *)

exception Contradiction of Proof.proof   (** a proof of False *)

exception Proved of Proof.proof          (** a proof of the current goal 
                                          * assumption *)


type ident = string                     (** The unique identifier of a 
                                         * satisfiability procedure *)

type message = 
    { body    : Logic.pred;             (** The predicate being asserted. This 
                                         * is most often an atom, but it 
                                         * might be the left-hand side of an 
                                         * implication, when that is not a 
                                         * conjunction. *)
      owner   : ident;                  (** Who is responsible for producing a 
                                         * proof of this  *)
      id      : int;                    (** The responsible satisf. proc can 
                                         * use this to later identify the 
                                         * message  *)
      goal    : bool;                   (** If it is a goal. In this case the 
                                         * asserted predicate is the negation 
                                         * of "body"  *)
      aboutToFail: bool;                (** If this flag is set in a message, 
                                         * it means that, unless there is a 
                                         * contradiction for this message, we 
                                         * are going to fail the proof. The 
                                         * sat. procedures can use more 
                                         * expensive mechanisms if this is 
                                         * the case *)
    } 
      

                                        (** There are the entry points in a 
                                         * decision procedure *)
type entries = 
    {
                                        (** Initialize.  *)
      init:       unit -> unit;
                                        (** Record a snapshot *)
      snapshot  : unit -> unit;
                                        (** Discard all assertions since last 
                                         * not-undone snapshot *)
      undo      : unit -> unit;
                                        (** Assert a literal (the second 
                                         * argument). Accumulate the list of 
                                         * equalities to the first argument. 
                                         * Might raise Contradiction, or (if 
                                         * the assertion is a goal) Proved  *)
      assume    : message list -> message -> message list;
                                        (** Prove a message generated earlier.*)
      proveIt   : message -> Proof.proof;

        
      debug     : bool ref;             (* For setting a debug flag *)

      ident     : ident;                (** The identifier *)
                
      descr   : string;                 (** A short description *)
    } 


                               (** There is a global pointer to a procedure 
                                * for proof generation. satisfiability 
                                * procedures can call this when they need to 
                                * produce a proof for a message belonging to 
                                * some _OTHER_ satproc.  *)

let proveIt : (message -> Proof.proof) ref = 
  ref (fun _ -> E.s (E.bug "proveIt is not defined"))


  (** Pretty print a message *)
let d_message () msg = 
  dprintf "%s(%s:%a)"
    (if msg.goal then "Goal" else "Assert")
    msg.owner
    L.d_pred msg.body

module type SATPROC = 
  sig
  
                                        (** Called to initialize the decision 
                                         * procedure and to create a table of 
                                         * the entry points. *)
    val initialize : ident -> entries
  end



(** A specification for a parser *)
type parserSpec = {
    language: string; (** The name of the language. This is used to construct
                       * a command line option --langinput *)
    initParser: string -> Lexing.lexbuf -> unit; (** Initialization *)
    nextFewInputs: Lexing.lexbuf -> Input.input list;
  } 
