
(* Kettle Graphical User Interface
 * Original Implemented with LablTk ToolKit, by Kun Gao
 * Adapted to Kettle by George Necula
 *)

(* --- Supporting Code --- *)
open Logic            (* ntermediate language *)
open Options
open Pretty           (* pretty printing *)
open Tk               (* open the Tk interface *)


(* --- Support Modules --- *)
module E = Errormsg   (* Error messaging module *)
module G = Engine     (* Engine definitions *)
module ST  = Stree    (* State tree *)
module H = Hashtbl    (* hashtable *)
module GL = Guilib
module SW = Sourcewin


module U = Util
module GA = GrowArray
module P = Proof

(* --- Type declarations --- *)
type tree = ST.tree               (* tree type *)
type treeNode = ST.treeNode       (* node of tree *)
type canvas = Widget.canvas Widget.widget     (* canvas type *)
type toplevel = Widget.toplevel Widget.widget (* toplevel type *)
type frame = Widget.frame Widget.widget       (* frames *)
type regGroup = { mutable gName: string;      (* name of group *)
		  mutable gRegs: U.registerInfo list; (* registers in group *)
		  mutable gChange: int list;  (* registers that changed *)
		}


let showSource = ref true

let sourceColor = ref "black"

(* --- Gui Global variables and Flags --- *)

(* string of current source file, empty string means no reloading *)
let nextSrcName: string ref = ref ""

(* are we finished or not *)
let finished: bool ref = ref false

(* this keeps track of whether we have stopped or not *)
let isStopped: bool ref = ref false  (* initially we are not stopped *)

(* Breakpoints *)
let breakPts: int list ref = ref []

(* Global Hash tables used by Src Sal Class *)
let lineMapped: (int, bool) H.t = H.create 113
let srcLineMap: (int, int) H.t = H.create 113

(* Global Register tracking used by Registers *)
(* a list of all active register groups *)
let regGroups: (string * frame) list ref = ref []

(* --- GUI Options Variable --- *)
let breakOnErrors: bool ref = ref true
let stepBackNum: int ref = ref 0

(* ----- Class Interface for TK GUI ----- *)
(* UI is composed of components managed by a main window. *)
(* Each component knows how to initialize and update itself *)
(* when given the state tree. It is up to the main manager window *)
(* to place each component in the correct location. *)
(* furthermore, the manager window provide services that *)
(* each component can utilize *)


let options : optionDescr list =
  SW.sourcewinOptions	

type kettleState = {
    (** The goals in this block, in reverse order. Those without proofs are 
     * either failed or not yet finished *)
    mutable goals: (pred * P.proof option) list;

    (** The assumptions in this block, in reverse order *)
    mutable assumes: (U.symbol * pred) list;


    (** Whether there are any not yet finished proofs, or in nested blocks *)
    mutable hasUnfinishedProofs: bool;
  }


exception IsKettleState of kettleState
let castToKettleState (state: ST.state) : kettleState = 
(*
  ignore (E.log "Calling cast for node %s:%s\n" 
            state#name state#infoline);
*)
  try state#downcast;
      E.s (E.error "downcast did not raise an Exception")
  with IsKettleState s -> s
    | _ -> E.s (E.error "Failure casting state to Kettle state")

class kstate (assumptions: (U.symbol * pred) list)
             (blockpars: (string * string) list) = 
  let s = { goals = []; assumes = []; hasUnfinishedProofs = false } in
object (self)
  val s = s 
  method private getParam (n: string) : string = 
    match List.filter (fun (n', _) -> n' = n) blockpars with 
      (_, what) :: _ -> what
    | [] -> raise Not_found

  method name = 
    try self#getParam "name" with Not_found -> ""

  method infoline = 
    self#name ^ 
    (match self#filename with 
      None -> ""
    | Some f -> " at " ^ f ^ 
        (match self#line with 
          None -> ""
        | Some l -> " line " ^ string_of_int l))

  method filename = 
    try Some (self#getParam "file") with Not_found -> None

  method line = 
    try Some (int_of_string (self#getParam "line")) with _ -> None

  method downcast = raise (IsKettleState s)

  method displayAssumptions = 
    List.map (fun (n, p) -> dprintf "%s: %a" (U.symbolName n) G.d_pred p) 
      assumptions


  method displayState = 
    (* We will display the goals *)
    let acc = ref [] in
    let id = ref 0 in
    List.iter (fun (g, po) -> 
      acc := 
        { U.rName = "goal" ^ string_of_int !id;
          U.rGroup = "Goals";
          U.rOneLineVal = None;
          U.rVal = G.d_pred () g } :: !acc;
      acc := 
        { U.rName = "proof" ^ string_of_int !id;
          U.rGroup = "Goals";
          U.rOneLineVal = None;
          U.rVal = 
            (match po with 
              None -> text "Failure"
            | Some p -> P.d_prf () p) } :: !acc;
      incr id)

      s.goals;

    List.iter (fun (n, p) -> 
      acc := { U.rName = U.symbolName n;
               U.rGroup = "Goals";
               U.rOneLineVal = None;
               U.rVal = dprintf "%a" G.d_pred p } :: !acc)
      s.assumes;

    List.rev ! acc

  method copy  = ({< s = { s with goals = s.goals} >} :> ST.state)

end  


class kettleSource
    (top: 'a Widget.widget) : GL.uiComponent
    = 

  (** We will load all source files into one stream, for now. *)
  let sourceInfo = GA.make 5000 (GA.Elem (SW.newInfoEntry ("", 0))) in

  (** Keep track of the loaded files, and the index of the first line *)
  let loadedFiles: (string, int) H.t = H.create 13 in
  let nextFreeIndex = ref 1 in

  let srcTag = "s" in

  (* A function to insert source *)
  let insertSrcLine        (* Now save the configuration *)
      _
      ((str: string), (lno: int)) 
      (txt: Widget.text Widget.widget) 
      (idx: textIndex) =
    let str = str ^ "\n" in
    (* insert the text *)
    Text.insert txt ~tags:[srcTag] ~index:idx
      ~text:(string_of_int lno ^ ":\t" ^ str);
    1  (* Always one line *)
  in                       

  (** Make a stream out of it *)
  let srcStrm = 
    new SW.sourceStreamFromInfo
      ~name:"Source"
      ~show:showSource
      ~insertInstr:insertSrcLine
      ~info:sourceInfo 
  in

  object (self)
    inherit SW.sourceWindow ~saveID:"sourcecode" top
    method name = "Source code"


    method update (force: bool) = begin
      let nd = GL.currNode () in
      (match nd.ST.state#filename with 
        Some f -> 
          self#setFileName f;
          (match nd.ST.state#line with 
            Some l -> self#setLineNumber l
          | _ -> ())

      | _ -> ())
    end

    (** Set the current file name *)
    val mutable fileStartIndex = 0
    method private setFileName (f: string) = 
      (** See if we can find this file *)
      let startlno = 
        try H.find loadedFiles f 
        with Not_found -> begin
          let idx = !nextFreeIndex in
          (* Add one line with the file name *)
          GA.setg sourceInfo idx
            (SW.newInfoEntry ("****** File " ^ f ^ " ******", 0));
          nextFreeIndex := 
            SW.insertFileIntoSourceInfo sourceInfo (idx + 1) f;
          H.add loadedFiles f (idx + 1);

          (* We do not set the sup and the inf because we use only one stream 
           * for now *)
          self#setSources [ srcStrm ] 0 (fun () -> ());
          idx + 1
        end
      in
      fileStartIndex <- startlno - 1


      

    method private setLineNumber (l: int) = 
      let l' = fileStartIndex + l in
      if l' > 0 && l' <= GA.max_init_index sourceInfo then 
        self#selectLine 0 l'
      else
        ignore (E.warn "Trying to select line number %d\n" l)

      
    method init = 
      fileStartIndex <- 0;
      nextFreeIndex := 1;

        (* Color the display *)
      let font = "Helvetica -12" in
      
      Text.tag_configure self#textbox 
        ~tag:srcTag 
        ~foreground:(`Color !sourceColor)
        ~font:font;
      
      H.clear loadedFiles;
      GA.reset_max_init_index sourceInfo

end


(* --- Class For Menu System --- *)
class menu (top: 'a Widget.widget) : GL.uiComponent =
  (* widgets used for menu *)
  let mbar = Frame.create ~relief:`Raised top in
  (* menu items *)
  let m1 = Menubutton.create ~text:"File" ~underline:0 mbar in
  let m3 = Menubutton.create ~text:"Options" ~underline:0 mbar in
  let m4 = Menubutton.create ~text:"Registers" ~underline:0 mbar in
  let m5 = Menubutton.create ~text:"Help" ~underline:0 mbar in
  let group_menu = Menu.create m4 in
  let _ = (* pack the main menu items *)
    pack [m1; m3; m4] ~side:`Left;
    pack [m5] ~side:`Right;
    Menubutton.configure ~menu:group_menu m4;
  in
  
  object (self)
    inherit GL.defaultUIComponent ~name:"Menu"
                                  ~expand:false
                                  ~fill:`X
                                  ~dockable:false
                                  ~saveID:""
                                  mbar

    method copy (tl: toplevel) = 
      new menu tl

    (* Function that takes a menu, and returns a function
       while takes one option, and places it in the menu *)
    method private mkOption (menu: Widget.menu Widget.widget)
	(opt: optionDescr) = 
      GL.mkOption top menu opt 
        (fun () -> ())


    (* --- My Public Methods --- *)
    method init = begin (* init menus *)
      (* File Menu *)
      let m1a = Menu.create m1 in
      Menubutton.configure ~menu:m1a m1;
      
      (* -- separator -- *)
      Menu.add_separator m1a;

      (* menu options - quit command *)
      Menu.add_command ~label:"Quit" 
	~command:GL.closeUI 
        m1a;


      (* The UI Option Submenu *)
      let gui_options = Menu.create m3 in
      List.iter
	(self#mkOption gui_options)
	options;

      (* Options Menu *)
      (* We add one item for the architecture and one for the vcgen. For now we 
       * put in there only those things that can be changed without restarting *)
      let option_menu = Menu.create m3 in
      
      (* add GUI submenu *)
      Menu.add_cascade ~label:"GUI Options" ~underline:0
	~menu:gui_options option_menu;

      Menubutton.configure ~menu:option_menu m3;

    end

    (* update *)
    method update (force: bool) = begin
      ()
    end
  end

(** A class to customize the GUI *)
class kettleCustomGui = object

    (* This is used for showing the node label *)
    method showNodeLabel (nd: ST.treeNode) (can: GL.canvas) 
                         ~(x:int) ~(y:int) = 
      let ks = castToKettleState nd.ST.state in
      let hasChildren = nd.ST.children <> [] in
      (* Now add the label *)
      let newText = 
	Canvas.create_text can 
          ~x:x ~y:y
          ~anchor:`W
          ~fill:
          (if nd == ST.currNode (GL.theTree ()) then `Blue 
          else if ks.hasUnfinishedProofs then `Red
          else `Black)
          ~text: nd.ST.state#name;
      in
      (* For nodes that are not done, and are leaves, we want to be able 
       * to select them *)
      Canvas.bind 
        ~events:[`ButtonPress] 
        ~fields:[`MouseX; `MouseY]
        ~action:
        (fun _ -> 
          ST.setCurrent (GL.theTree ()) nd;
          GL.updateUI true)
        can newText;

      newText
      
end

let run () : unit =
  ignore (E.log "Initialize VCGen\n");

  let configFile = "kettle_gui.cfg" in
  
  (* Load the configuration *)
  U.loadConfiguration configFile;

  U.useConfigurationInt "updateInterval" (fun i -> GL.updateInterval := i);

  U.useConfigurationList "sourceDirectories"
    (fun l -> 
      SW.searchDirectories := 
        List.map 
          (fun s -> 
            match s with 
              U.ConfString s -> s
            | _ -> 
                ignore (E.warn 
                          "Found invalid configuration in searchDirectories");
                "")
          l);


  GL.initGui ();
  
  
  (* Configure the Main and Execution TK Window *) 
  Wm.title_set (GL.tkTop ()) "Visual VCGen Interface";
  
  let top = GL.tkTop () in

  let menu: GL.uiComponent = new menu top in
  
  (* instantiate source *)
  let src_view = new kettleSource top in


  (* instantiate assumptions *)
  let assumptions = 
    new GL.registers 
      ~title:"Assumptions"
      ~saveID:"Assumptions"
      ~popupTitle:(fun ri -> "Assumption")
      ~getState:(fun _ -> 
        List.map
          (fun d -> 
            { U.rName = ""; U.rGroup = "";
              U.rVal = d; U.rOneLineVal = None })
          (GL.currNode ()).ST.state#displayAssumptions)
      top 
  in

  let stateTree = new GL.stateTreeGui top (new kettleCustomGui) in

  (* instantiate infobar *)
  let status = new GL.info_line top in
  
  
  (* Reset my settings *)
  nextSrcName := "";
  finished := false;
  isStopped := false;
  breakPts := [];
  regGroups := [];
  
  (* instantiate groups *)
  let goals = 
    new GL.registers 
      ~title:"Goals"
      ~saveID:"Goals"
      ~getState:(fun _ -> (GL.currNode ()).ST.state#displayState) 
      top
  in

  (* make a container for everything *)
  let cont = new GL.container top in
  (* Must pack the container frame *)
  pack [cont#frame] ~inside:top ~expand:true ~fill:`Both;

  cont#setPanes
    (`Col [ `Win menu;
            `Row [ `Col [ `Win stateTree;
                          `Win goals ];
                   `Col [ `Win src_view;
                          `Win assumptions; ] ];
            `Win status ]);

  cont#init; (* Init the whole container *)
  
  GL.theUIComponents := [ (cont :> GL.uiComponent) ];

  GL.updateUI true; (* update UI for the first time *)
  
  (* ----- Invoke the Main Loop ----- *)
  mainLoop();
  
  (* Save some configuration *)
  U.setConfiguration "updateInterval" (U.ConfInt !GL.updateInterval);
  U.setConfiguration "sourceDirectories" 
    (U.ConfList (List.map (fun s -> U.ConfString s) !SW.searchDirectories));

  (* Now save the configuration *)
  U.saveConfiguration configFile;
  

  ()


(* ----- Interface to G.uiType ----- *)
class ui : G.uiType = 
  (* we make a new gtkui object *)

  (* make the new execution tree *)
  let _ = GL.initTree [] in
  let currNode : ST.treeNode ref = ref (ST.root (GL.theTree())) in

  object (self)
    method name = "gui"
    method options = options
    method enterBlock (assumptions: (U.symbol * pred) list) 
                      (params: (string * string) list) = 
      let t = GL.theTree () in
(*
      ignore (E.log "Entering block %a with %d assumptions\n"
                (docList (chr ';') (fun (n,p) -> dprintf "%s=%s" n p))
                params
                (List.length assumptions));
*)
      currNode := ST.newNode t (new kstate assumptions params) !currNode

    method exitBlock = 
      let p = !currNode in 
      if p.ST.parent = None then 
        E.s (E.error "Gui.exitBlock called outside any block!");

      (* If this block does not have unfinished proofs, we collapse it *)
      let thiss = castToKettleState p.ST.state in
      if not thiss.hasUnfinishedProofs then 
        p.ST.collapsed <- true;

      (* Set the new current node to the parent of this one. *)
      currNode := 
         (match p.ST.parent with 
           None -> E.s (E.bug "Found exit block at top level")
         | Some p' -> p')


    method private getCurrentState = 
      castToKettleState (!currNode).ST.state
      
    method start_goal (g: pred) = 
      if !currNode.ST.parent = None then 
        E.s (E.error "Gui.start_goal called outside any block!");

      (* Add this goal to the current block *)
      let s = self#getCurrentState in
      s.goals <- (g, None) :: s.goals

    method goal_success (g: pred) (p: P.proof) = 
      if !currNode.ST.parent = None then 
        E.s (E.error "Gui.goal_success called outside any block!");

      (* If the proof is "invalid" then it does not count *)
      match p with 
        P.G("invalid", []) -> self#goal_failure g
      | _ -> begin
          let s = self#getCurrentState in
          match s.goals with 
            (_, None) :: rest -> s.goals <- (g, Some p) :: rest
          | _ -> E.s (E.bug "goal_success but no goal_start")
      end

    method goal_failure (g: pred) = 
      if !currNode.ST.parent = None then 
        E.s (E.error "Gui.goal_failure called outside any block!");
      (* mark this node and all parents with hasUnfinishedProofs *)
      let p = !currNode in
      let thiss = castToKettleState p.ST.state in
      thiss.hasUnfinishedProofs <- true;
      
      ST.iterParents 
        (fun p -> 
          let ps = castToKettleState p.ST.state in 
          ps.hasUnfinishedProofs <- true)
        p

    method assume (name: U.symbol) (p: pred) = 
      if !currNode.ST.parent = None then 
        E.s (E.error "Gui.assume called outside any block!");
      let s = self#getCurrentState in
      s.assumes <- (name, p) :: s.assumes

    method unassume = 
      ()


    method run = 

      (* Set the current node *)
      let t = GL.theTree () in
      (match (ST.root t).ST.children with 
        [] -> ignore (E.warn "There are no root nodes to show\n")
      | r :: _ -> ST.setCurrent t r; 
          run ());

      (* We might have opened some blocks but have not closed them. This 
       * means that the *)
      ()
  end
