(* A library of functions for the VCGen Graphical User Interface
 * Implemented with LablTk ToolKit
 * COOLAID Project: 2003
 *)

(* --- Supporting Code --- *)
open Options
open Pretty           (* pretty printing *)
open Tk               (* open the Tk interface *)


(* --- Support Modules --- *)
module E = Errormsg   (* Error messaging module *)
module H = Hashtbl    (* hashtable *)
module GA = GrowArray
module U = Util
module GL = Guilib
module G = Engine

type frame = Widget.frame Widget.widget       (* frames *)
type widget = Widget.any Widget.widget
type label = Widget.label Widget.widget

type toplevel = Widget.toplevel Widget.widget (* toplevel type *)


let searchDirectories: string list ref = ref [ "." ]

let rec searchFile (fname: string) : in_channel = 
  let rec loop (dirs: string list) = 
    match dirs with 
      [] -> raise Not_found
    | d :: rest ->
        try
          let cin = open_in (d ^ "/" ^ fname) in
          cin
        with Sys_error _ -> loop rest
  in
  (* Try it first with no search directories *)
  try open_in fname
  with Sys_error _ -> loop (!searchDirectories)


let sourcewinOptions = 
  GL.guiOptions @
  [    
    { optInUI   = "Search &Directories";
      optRestart = false;
      optKind   = OStringList (';', searchDirectories);
      optExtra  = (fun () -> ());
      optCommandLine = "-searchDirs";
      optHelp = "Search directories (; separated)"; };
       
    { optInUI   = "&Reset Search Directories";
      optRestart = false;
      optKind   = OUnit;
      optExtra  = (fun () -> searchDirectories := []);
      optCommandLine = "-resetSearchDirs";
      optHelp = "Reset search directories"; };
  ] 
    


type sourceStream = < 

    sourceName: string;
    (** A name that differentiates this stream from other streams *)

    show: bool ref;    
    (** Whether this view is enabled *)

    nrInstrs: int;
    (** How many instructions *)

    getInstrSupLine: int -> int;
    getInstrInfLines: int -> lineRange list;

    getInstrVirtualLines: int -> int * int;
    setInstrVirtualLines: int -> int * int -> unit;
    (** Read one instruction information *)
 
    insertInstr: int -> Widget.text Widget.widget -> textIndex -> int;
    (** For each instruction, insert it into the textbox and return the 
     * number of window lines it takes *)
>

and 'a instrInfo = { 
    mutable data: 'a;  (** Carry some data for each instruction *)

    mutable supLine: int;
    (** The line in the super-source that it corresponds to. May be 0, to 
     * denote that it does not correspond to a super-line *)

    mutable infLines: lineRange list;
    (** An ordered list of ranges of inferior lines that this corresponds to. 
     * May be empty, to indicate that there is not inferior instruction for 
     * this one. *)

    mutable virtualLines: int * int
    (* The starting virtual lines where this is shown, and the number of 
     * lines. This is only valid for the source streams that are displayed. *)
  }

and lineRange = {
    mutable rFirst: int; 
    mutable rLast: int; 
  }

let dumpRanges (s: 'a instrInfo GA.t) (dumpi: 'a -> string) = 
  for i = 0 to GA.max_init_index s do 
    let ii = GA.get s i in
    ignore (E.log "%03d %s:\n\tsup:%d inf:%a\n"
              i 
              (dumpi ii.data)
              ii.supLine
              (docList
                 (fun r -> dprintf "%d-%d" r.rFirst r.rLast))
              ii.infLines)
  done

class sourceStreamFromInfo 
    ~(name: string)
    ~(show: bool ref)
    ~(insertInstr: 
        int -> 'a -> Widget.text Widget.widget -> textIndex -> int)
    ~(info: 'a instrInfo GA.t) = 
  object
    method sourceName = name
    method show = show
    method nrInstrs = (GA.max_init_index info) + 1
    method getInstrSupLine (i: int) = 
      (GA.get info i).supLine
    method getInstrInfLines (i: int) = 
      (GA.get info i).infLines
    method getInstrVirtualLines (i: int) = 
      (GA.get info i).virtualLines
    method setInstrVirtualLines (i: int) (l: int * int) = 
      (GA.get info i).virtualLines <- l

    method insertInstr (i: int) (txt: Widget.text Widget.widget) 
                       (where: textIndex) : int =  
      insertInstr i (GA.get info i).data txt where

  end
    
(** A function that fills in the inferior ranges *)
let fillInInferiorRanges (s: 'a instrInfo GA.t) 
                         ~(inf: 'b instrInfo GA.t) : unit = 
  (* Scan the inferior array and if we find ranges of instructions that have no 
   * superior line, but which are bordered by instructions with identical 
   * superior line, then assign them all the same superior line *)
  let lastSupLine = ref (-1) in
  for i = 0 to GA.max_init_index inf do 
    let ii = GA.get inf i in
    if ii.supLine > 0 then begin
      if ii.supLine = !lastSupLine then begin (* go back and fill *)
        let rec fill (j: int) = 
          if j >= 0 && (GA.get inf j).supLine = 0 then begin
            (GA.get inf j).supLine <- !lastSupLine;
            fill (j - 1)
          end
        in
        fill (i - 1)
      end else (* No the same one *)
        lastSupLine := ii.supLine
    end
  done;


  (* Scan the inferior array, and construct the ranges in reverse order. We 
   * do not do the 0th invalid instruction. *)
  for i = 1 to GA.max_init_index inf do 
    let ii = GA.get inf i in
    let l = ii.supLine in
    (* Insert [i] in the given range for [l] , which is reverse order. 
     * Precondition: i is less than the start of the previous range. *)
    let rec insertRange (r: lineRange list) : lineRange list = 
      match r with 
        [] -> (* We must add a new element *) [ { rFirst=i; rLast=i} ]
      | h :: rest ->
          (* We never have duplicates *)
          assert (i < h.rFirst || i > h.rLast);
          if i = h.rLast + 1 then begin
            h.rLast <- i; (* Increase the current range *)
            r
          end else if i > h.rLast then begin
            (* We add a new range *)
            { rFirst = i; rLast = i} :: r
          end else if i = h.rFirst - 1 then begin 
            (* We tack it to the start of this one *)
            h.rFirst <- i;
            (* See if we need to collapse the range *)
            (match rest with 
              h' :: rest' when h'.rLast = i - 1 -> 
                h'.rLast <- h.rLast;
                (* Drop h *)
                rest

            | _ -> r)
          end else if i < h.rFirst then begin
            (* We must recurse *)
            let rest' = insertRange rest in
            if rest != rest' then h :: rest' else  r
          end else begin
            E.s (E.bug "fillInInferiorRanges")
          end
    in
    assert (l >= 0 && l <= (GA.max_init_index s));
    let infLines = (GA.get s l).infLines in
    let infLines' = insertRange infLines in
    if infLines' != infLines then 
      (GA.get s l).infLines <- infLines'
  done;

  (* Now revert the lists *)
  for j = 0 to GA.max_init_index s do 
    match (GA.get s j).infLines with 
      [] | [ _ ] -> ()
    | l -> (GA.get s j).infLines <- List.rev (GA.get s j).infLines
  done;

  ()

  (* Make a new info entry *)
let newInfoEntry (data: 'a) : 'a instrInfo = 
  { data = data; supLine = 0; 
    infLines = [];
    virtualLines = (-1, 1) }
        


(** Insert a file into a source stream at a given index. Return the next 
 * usable index. *)
let insertFileIntoSourceInfo 
                       (instrs: (string * int) instrInfo GA.t) 
                       (start:int) 
                       (fname: string) : int = 
  let index = ref start in
  try 
    let cin = searchFile fname in
    (try
      let rec loop () = 
        let str = input_line cin in
        (* Add the line to the array *)
        GA.setg instrs !index (newInfoEntry (str, !index - start + 1));
        incr index;
        loop () 
      in
      loop ()
    with End_of_file -> close_in cin);

    (* Add one more line. The parser sometimes uses a line number that is one 
     * too big *)
    GA.setg instrs !index (newInfoEntry ("", !index -  start + 1));
    incr index;

    !index
    
  with Not_found -> 
    ignore (E.warn "Cannot open file %s. CWD=%s, Search path is %a" 
              fname (Sys.getcwd ()) 
              (docList ~sep:(chr ';') text) !searchDirectories);
   start
      



(** Iterate over some ranges *)
let rec iterRanges (f: int -> unit) (r: lineRange list) = 
  match r with 
    [] -> ()
  | r :: rest -> 
      for j = r.rFirst to r.rLast do 
        f j
      done;
      iterRanges f rest

(** Map a window line to a source stream index, and an index within the 
 * source stream *)
type instrPosition = int * int 

let virtualLineToInstrPosition : (int, instrPosition) H.t = H.create 113

(* We display the source code in pages. A page has a fixed number of "main" 
 * source lines. *)
type pageInfo = 
    NotLoaded of (unit -> unit) (* The page is not loaded, and this is the 
                                * loader thunk *)
  | Loaded of pageData

and pageData = { 
    mutable pgWindowLine: int; (** The window line corresponding to the first 
                                * line in this page. *)
    mutable pgVirtualLine: int; (** The virtual line corresponding to the 
                                  * first line in this page. Virtual lines 
                                  * are used to map to instruction positions *)
    mutable pgNrLines: int; (** number of window/virtual lines in this page *)
    mutable pgMainStart: int; (** the first main instruction in this page *)
  }


let debugDisplay = false


class sourceWindow
    ?(title="Source code")
    ?(textwidth=60)
    ?(textheight=30)
    ?(expand=true)
    ?(fill=`Both)
    ?(pagingSize=100)
    ?(name="Source code")
    ?(saveID="source_code")
    (top: 'b Widget.widget) = 
      
  (* widgets used for source and sal display *)
  let srcframe = Frame.create ~borderwidth:1 top in
  let title = Label.create ~text:title srcframe in
  let srccode = Text.create ~width:textwidth ~state:`Normal
                            ~height:textheight srcframe in
  let srccodescrl_h = Scrollbar.create ~orient:`Horizontal srcframe in
  let srccodescrl_v = Scrollbar.create ~orient:`Vertical srcframe in

  (** an array of pages. Page 0 is reserved for the instruction that do not 
   * have a main. Page 1 is for the main instructions 1...pagingSize, ... *)
  let initPageData = NotLoaded (fun _ -> E.s (E.bug "pages are empty")) in
  let pages = GA.make 100 (GA.Elem initPageData) in

  (* Keep track of how many more pages we have to load. *)
  let unloadedPages = ref 1 in
  
  (** Estimate the total number of virtual lines. Each unloaded page counts 
   * as pagingSize (we only count the main lines in it) *)
  let unloadedPageSizeEstimate = pagingSize in
  let virtualLinesEstimate = ref 0 in


  (** The current window line *)
  let currVirtualLine = ref 1 in

  let initPages () = 
    unloadedPages := 1;
    virtualLinesEstimate := 1;

    if debugDisplay then 
      ignore (E.log "Sourcewin: initializing pages\n");
    GA.clear pages; (* Like nothing was written *)
  in
  (** Map a window line to a page number, the page data for that page, and an 
   * estimate of the virtual line number, where the unloaded pages are 
   * estimated to contain unloadedPageSizeEstimate *)
  let windowLineToPageNumber (wl: int) : int * pageData * int = 
    let rec loop (est_sofar: int) (i: int) = 
      (* This page must be loaded *)
      if i > GA.max_init_index pages then
          E.s (E.error "windowLineToPageNumber: wl=%d" wl);
      match GA.get pages i with 
        NotLoaded _ -> loop (est_sofar + unloadedPageSizeEstimate) (i + 1)
      | Loaded pd -> 
          if pd.pgWindowLine <= wl && wl < pd.pgWindowLine + pd.pgNrLines then 
            i, pd, est_sofar + (wl - pd.pgWindowLine)
          else
            loop (est_sofar + pd.pgNrLines) (i + 1)
    in
    loop 1 0
  in

  let mainLineToPageNumber (mainl: int) : int = 
    1 + (mainl - 1) / pagingSize
  in
    
  (* Page in main line *)
  let pageInMainLine (l: int) : unit = 
    let pageno = mainLineToPageNumber l in
    let pageoff = (l - 1) mod pagingSize in
    if pageno < 0 || pageno > GA.max_init_index pages then 
      () (* Outside our range *)
    else begin
      if debugDisplay then 
        ignore (E.log "Sourcewin: paging in main=%d, page=%d, offset=%d\n"
                  l pageno pageoff);
      match GA.get pages pageno with 
        Loaded pd -> 
          if debugDisplay then 
            ignore (E.log "  already loaded windowline=%d\n"
                      (pd.pgWindowLine + pageoff))
            
      | NotLoaded thunk -> begin
          if debugDisplay then 
            ignore (E.log "  loading ...\n");
          thunk ();
          match GA.get pages pageno with 
            Loaded pd -> 
              if debugDisplay then 
                ignore (E.log "  finished loading windowline=%d\n"
                          (pd.pgWindowLine + pageoff))

          | NotLoaded _ -> E.s (E.bug "Loading of page did not succeed")
      end
    end
  in
  (** Map a window line to a virtual line *)
  let windowLineToVirtualLine (wl: int) = 
    let _, pd, _ = windowLineToPageNumber wl in
    pd.pgVirtualLine + (wl - pd.pgWindowLine)
  in

  (** Compute a window line, based on a fraction from the start *)
  let windowFractionToLine (frac: float) = 
      1 + int_of_float (0.5 +. (float_of_int !currVirtualLine) *. frac) 
  in
    
  let virtualFractionToLine (frac: float) = 
      1 + int_of_float (0.5 +. (float_of_int !virtualLinesEstimate) *. frac) 
  in

  (** The command that is invoked by the text box when the view changes. It 
   * should set the scrollbar visible portion. It might also have to reload 
   * some text, if we have changed the text because we pressed some keys *)
  let yscrollcommand ~(first:float) ~(last:float) = 
    (* The parameters are based on the number of lines loaded into the text 
     * box (!currVirtualLine). We want them based on the estimate of lines we 
     * will eventually have (!virtualLinesEstimate). Window lines are 1 
     * based. *)
    let window_yfirst = windowFractionToLine first in
    let window_ylast = windowFractionToLine last - 1 in

    if debugDisplay then 
      ignore (E.log "yscrollcommand: window_yfirst=%d, window_ylast=%d\n"
                     window_yfirst window_ylast);

    (* Sometimes yscrollcommand is called with window_yfirst=1 before the 
     * windows are actually intialized *)
    if window_yfirst > window_ylast ||
       window_yfirst = 1 && window_ylast = 1 then 
      
      ()
    else begin
      (* Estimate the virtual line of the first window_line *)
      let _, _, est_yfirst_virtual = windowLineToPageNumber window_yfirst in
      
      Scrollbar.set srccodescrl_v 
        ~first:((float_of_int est_yfirst_virtual) /. 
                (float_of_int !virtualLinesEstimate))
        ~last:((float_of_int (est_yfirst_virtual + 
                                window_ylast - window_yfirst))
                 /. (float_of_int !virtualLinesEstimate));

     (* See if we have to load some more text. This happens if the visible 
      * portion contains a gap, or if there is a gap immediately before the 
      * first line, or immediately after the last line. *)
    end
  in

  let yscrollcommand_wrapper ~(first:float) ~(last:float) = 
    if !unloadedPages = 0 then 
      Scrollbar.set srccodescrl_v ~first:first ~last:last
    else
      yscrollcommand  ~first:first ~last:last
  in

  (** The command that is invoked by the scrollbar when we scroll *)
  let yscroll ~(scroll: scrollValue) = 
    (* The values from the text box are based on currVirtualLine. *)
    let yfirst, ylast = Text.yview_get srccode in
    (* Compute the window first and last lines. They are 1-based. *)
    let window_yfirst = windowFractionToLine yfirst in
    let window_ylast = windowFractionToLine ylast - 1 in

    (* Compute the window page size *)
    let page_size = window_ylast - window_yfirst + 1 in

    let yfirst_page, yfirst_pd, yfirst_est_virtual = 
      windowLineToPageNumber window_yfirst in

    let yfirst_offset = window_yfirst - yfirst_pd.pgWindowLine in

    (* Given a page and an offset from its start, make sure to show that 
     * line. pageStart=true means that this is the start of a page, and we 
     * must also show the end as well *)
    let rec pageIn (curr_page: int) 
                   (to_scroll: int) (* how much from the start of the current 
                                     * page to the line that we care about *)
                   (to_scroll_from_start: bool) (* Whether the "to_scroll" is 
                                                 * from the start or from the 
                                                 * end of the page *)
                   (pageStart: bool)
                   : int = 
      if debugDisplay then 
        ignore (E.log "   scrolly%s: pageIn page=%d toscroll=%d%s\n" 
                  (if pageStart then "" else " (the end) ")
                  curr_page to_scroll
                  (if to_scroll_from_start then "" else " (from end)"));
      match GA.get pages curr_page with 
        NotLoaded thunk -> 
          thunk (); 
          pageIn curr_page to_scroll to_scroll_from_start pageStart
      | Loaded pd -> 
          let to_scroll = 
            if to_scroll_from_start then to_scroll 
            else to_scroll + pd.pgNrLines
          in
          if to_scroll >= 0 then (* We go forward *)
            if to_scroll < pd.pgNrLines 
                  || curr_page >= GA.max_init_index pages then begin
              (* This is the page for the first line. See if we must page in 
               * the page end as well *)
              if pageStart && 
                 to_scroll + page_size >= pd.pgNrLines &&
                 curr_page < GA.max_init_index pages then 
                ignore (pageIn (curr_page + 1) 
                               (to_scroll + page_size - pd.pgNrLines)
                               true
                               false);
              pd.pgWindowLine + to_scroll
            end else 
              pageIn (curr_page + 1) (to_scroll - pd.pgNrLines) true pageStart
          else (* We must go back *)
            if curr_page = 0 then 
              1
            else
              pageIn (curr_page - 1) to_scroll false pageStart
    in

    if debugDisplay then 
      ignore (E.log "yscroll: win_yfirst=%d, virt_yfirst=%d, page_yfirst=%d\n"
                window_yfirst yfirst_est_virtual yfirst_page);
    let new_window_yfirst = 
      match scroll with 
        `Unit n -> 
          if debugDisplay then 
            ignore (E.log "   scroll %d units\n" n);
          pageIn yfirst_page (yfirst_offset + n) true true
            
      | `Page n -> 
          if debugDisplay then 
            ignore (E.log "   scroll %d pages\n" n);
          pageIn yfirst_page (yfirst_offset + n * (page_size - 2)) true true
            
      | `Moveto f -> 
          (* Try to estimate the page where we land *)
          let new_yfirst = virtualFractionToLine f in

          (* Find the page in which new_yfirst lives, and its virtual line of 
           * the start of page *)
          let rec findStartPage (sofar: int) (page: int) : int * int = 
            let thisPageSize = 
              match GA.get pages page with
                NotLoaded _ -> unloadedPageSizeEstimate
              | Loaded pd -> pd.pgNrLines
            in
            if sofar + thisPageSize > new_yfirst 
                       || page >= GA.max_init_index pages then 
              page, sofar
            else
              findStartPage (sofar + thisPageSize) (page + 1)
          in
          let firstPage, firstPage_start_virtual = findStartPage 1 0 in

          if debugDisplay then 
            ignore (E.log "   scroll to new_yfirst=%d (%d lines) in page %d\n" 
                   new_yfirst (new_yfirst - yfirst_est_virtual)
                   firstPage);
          
          pageIn firstPage (new_yfirst - firstPage_start_virtual) true true
    in

    (* The window lines might have change. Recompute the window_yfirst *)

    (* The values from the text box are based on currVirtualLine. *)
    let yfirst, _ = Text.yview_get srccode in
    (* Compute the window first and last lines. They are 1-based. *)
    let window_yfirst = windowFractionToLine yfirst in 

    if debugDisplay then 
      ignore (E.log "   new_window_yfirst = %d (current one %d)\n" 
                new_window_yfirst window_yfirst);

    Text.yview srccode
      ~scroll:(`Unit (new_window_yfirst - window_yfirst)); 
    (* This will call the yscrollcommand eventually, who will set the 
     * scrollbar *)
  in

  let yscroll_wrapper ~(scroll:scrollValue) : unit = 
    if !unloadedPages = 0 then 
      Text.yview srccode ~scroll:scroll
    else
      yscroll ~scroll:scroll
  in

  (* pack main widgets *)
  let _ =
    pack [title] ~side:`Top ~fill:`X ~expand:false;

    pack [srcframe] ~side:`Left ~fill:`Both ~expand:true;

    pack [srccodescrl_v] ~side:`Right ~fill:`Y;
    pack [srccodescrl_h] ~side:`Bottom ~fill:`X;
    pack [srccode] ~side:`Top ~fill:`Both ~expand:true;
    Text.configure 
      ~yscrollcommand:yscrollcommand_wrapper
      srccode;

    Scrollbar.configure 
      ~command:yscroll_wrapper
       srccodescrl_v;
    Text.configure ~xscrollcommand:(Scrollbar.set srccodescrl_h) srccode;
    Scrollbar.configure ~command:(Text.xview srccode) srccodescrl_h;

  in

  object (self: 'a)
    inherit GL.defaultUIComponent    
                                     ~expand:expand
                                     ~fill:fill
                                     ~name:name
                                     ~saveID:saveID
                                     srcframe

    method private textbox = srccode

    method copy (tl: toplevel) : GL.uiComponent = 
      E.s (E.unimp "sourceWindow.copy")

    val mutable sources: sourceStream array = [| |] 

    val mutable main_idx = -1

    method private sources = sources

    (** A new sourceWindow number of source streams, where the first one is 
     * the higher-level one *)
    method private setSources (src: sourceStream list) 
                              (main: int) (* Either 0 or 1 *)
                              (after: unit -> unit) = begin
      (* Change the cursor *)
      (*
      List.iter (fun ui -> 
        Frame.configure ui#frame ~cursor:(`Xcursor "clock"))
        !GL.theUIComponents;
      Tk.update ();
      *)
                    
      main_idx <- main;
      sources <- Array.of_list src;
      H.clear virtualLineToInstrPosition;
      initPages ();
      currVirtualLine := 1;                          
                                
      (* Clear everything, in case we are re-initializing *)
      Protocol.remove_callbacks srccode;
                                

      (* Intercept some keys and pretend we are manipulating the scrollbar *)
      Tk.bind
         ~events:[`KeyPress]
         ~extend:false  (* So that we do not do it twice *)                    
         ~fields:[`KeySymString]
         ~action:(fun (e: Tk.eventInfo) ->
           (* ignore (E.log "Pressed keysym %s\n" e.ev_KeySymString); *)
           Protocol.debug := false;
           match e.ev_KeySymString with 
             "Next" -> yscroll ~scroll:(`Page 1)
           | "Prior" -> yscroll ~scroll:(`Page (-1))
           | "Up" -> yscroll ~scroll:(`Unit (-1))
           | "Down" -> yscroll ~scroll:(`Unit 1)
           | "Home" -> yscroll ~scroll:(`Moveto 0.0)
           | "End" -> yscroll ~scroll:(`Moveto 1.0)
           | _ -> ()) 
      srccode;

        (* Turn on editing, so that we can delete and insert. *)
      Text.configure ~state:`Normal srccode;
        (* Delete everything *)
      Text.delete srccode ~start:(`Linechar (0,0), []) ~stop:(`End,[]);

      (* Delete all tags *)
      Text.tag_delete srccode (Text.tag_names srccode); 
                                
      assert (0 <= main && main <= 1);

      if Array.length sources <= main then 
        E.s (E.bug "Sourcewin.init: not enough source streams");

      (* Main sure that we always show the main level *)
      if not !(sources.(main)#show) then begin
        ignore (E.warn "Sourcewin: The main source has show=false");
        (sources.(main)#show) := true
      end;

      (* We have a global offset from the currVirtualLine to the windowline *)
      let virtualToWindow = ref 0 in

      (* Display a line at a certain position *)
      let rec displayLine (srcidx: int) (idx: int) = 
        let src: sourceStream = sources.(srcidx) in
        let startLine = !currVirtualLine in
        let insert_where = startLine + !virtualToWindow in

        if ! (src#show) then begin
          let nrLines = 
            src#insertInstr idx srccode 
              (`Linechar (insert_where, 0), []) in
          if nrLines > 0 then begin
            if debugDisplay then
              ignore (E.log "Display line %d:%d as line %d\n" 
                        srcidx idx insert_where);
            currVirtualLine := nrLines + !currVirtualLine;
            (* Remember what lines we put in that position *)
            for i = startLine to !currVirtualLine - 1 do 
              H.add virtualLineToInstrPosition i (srcidx, idx)
            done;
            src#setInstrVirtualLines idx (startLine, nrLines);
          end;
        end


      (** Display a line and its inferiors, recursively *)
      and displayLineWithInf (srcidx: int) (idx: int) = 
        assert (srcidx >= 0 && srcidx < Array.length sources);
        let src = sources.(srcidx) in
        displayLine srcidx idx;
        if srcidx + 1 < Array.length sources then
          displayRanges (srcidx + 1) (src#getInstrInfLines idx)
        
      and displayRanges   (srcidx': int) 
                          (r: lineRange list) = 
        iterRanges (fun j -> displayLineWithInf srcidx' j) r
      in
      
      (* We must display first the inferior lines with no SUP, starting with 
       * the lowest level. We do that by finding the infLines of the 0th 
       * element in the next superior stream. *)
      currVirtualLine := 1;
      virtualToWindow := 0;
      for i = Array.length sources - 1 downto main + 1 do 
        if debugDisplay then
          ignore (E.log "display orphans at level %d\n" i);
        let src = sources.(i - 1) in
        displayRanges i (src#getInstrInfLines 0)
      done;
      (* The orphan lines go all into page 0 *)
      GA.setg pages 0 (Loaded { pgWindowLine = 1;
                                 pgVirtualLine = 1;
                                 pgNrLines = (!currVirtualLine - 1);
                                 pgMainStart = 0; });
      if debugDisplay then 
        ignore (E.log "displayed %d orphans\n" (!currVirtualLine - 1));


      (* Now we set the rest of the pages *)
      let loadMainInstrs (pageno: int) : unit = 

        (* Turn on editing *)
        Text.configure ~state:`Normal srccode;

        (* Find the insertion place. Look for the previous loaded page. *)
        let rec lookback (i: int) = 
          match GA.get pages i with 
            Loaded pd -> 
              if debugDisplay then 
                ignore (E.log "Will load page %d at window line %d\n"
                          pageno (pd.pgWindowLine + pd.pgNrLines));
              pd.pgWindowLine + pd.pgNrLines
          | NotLoaded _ -> 
              if i = 0 then E.s (E.bug "Sourcewin: loadMainInstrs");
              lookback (i - 1)
        in
        let first_win_line = lookback (pageno - 1) in
        let first_virtual_line = !currVirtualLine in
        
        virtualToWindow := first_win_line - !currVirtualLine;

        let mainsrc = sources.(main) in
        let start = 1 + (pageno - 1) * pagingSize in
        let last = min (mainsrc#nrInstrs - 1) (start + pagingSize - 1) in
        if debugDisplay then
          ignore (E.log "display lines at main level %d (from %d to %d)\n" 
                    main start last);
        for j = start to last do 
          (* See if we must display some superior lines. We display a 
           * superior line if we have one and the previous top line does not 
           * have the same superior line. In that case we display at most 10 
           * superior lines, of which only the last has inferior lines. *)
          let sup = mainsrc#getInstrSupLine j in
          if main > 0 &&
            sup > 0 && (j = 1 || mainsrc#getInstrSupLine (j - 1) <> sup) then 
            begin
              let super = sources.(0) in
              (* find first to display *)
              let rec findFirst (idx: int) = 
                if idx = 1 || idx < sup - 10 then idx 
                else if super#getInstrInfLines (idx - 1) != [] then 
                  idx
                else
                  findFirst (idx - 1)
              in
              let first = findFirst sup in
              for i = first to sup do 
                displayLine 0 i
              done
            end;
          (* Now we display the main line, with inferiors *)
          displayLineWithInf main j
        done;

        (* Now set the current page information *)
        let nrLines = !currVirtualLine - first_virtual_line in
        GA.set pages pageno
          (Loaded { pgWindowLine = first_win_line;
                    pgVirtualLine = first_virtual_line;
                    pgNrLines = nrLines;
                    pgMainStart = start });
        if debugDisplay then 
          ignore (E.log "Finished loading page %d (window=%d,virtual=%d,nrLines=%d)\n"
                    pageno first_win_line first_virtual_line nrLines);
        decr unloadedPages;

        virtualLinesEstimate := 
          !virtualLinesEstimate - unloadedPageSizeEstimate + nrLines;
            
        if !unloadedPages = 0 then begin
          if debugDisplay then 
            ignore (E.log "Finished loading all pages\n");
        end;

        (* Now we must scan the following pages and incrment their window 
         * lines *)
        for k = pageno + 1 to GA.max_init_index pages do 
          match GA.get pages k with 
            Loaded pd -> 
              pd.pgWindowLine <- pd.pgWindowLine + nrLines
          | _ -> ()
        done;

        (* Turn off editing *)
        Text.configure ~state:`Disabled srccode;
                   
      in

      let nrMainInstrs = sources.(main)#nrInstrs in
      unloadedPages := (nrMainInstrs + pagingSize - 1) / pagingSize;

      for page = 1 to !unloadedPages do 
        GA.setg pages page (NotLoaded (fun () -> loadMainInstrs page));
      done;

      virtualLinesEstimate := 
        !currVirtualLine + 
          !unloadedPages * unloadedPageSizeEstimate;

      (* Turn off editing *)
      Text.configure ~state:`Disabled srccode;

      after ();
      ()
   end

   method private showLine (src: int) (idx: int) = 
     if src < 0 || src > Array.length sources || not !(sources.(src)#show) 
     then 
       E.s (E.bug "Sourcewin: viewLine invalid source %d\n" src);

     (* Find first the main instruction to which this one belongs *)
     let rec findMainLine (src': int) (idx': int) = 
       if src' = main_idx then 
         idx'
       else if src' < main_idx then begin
         (* Main is an inferior line *)
         if src' + 1 <> main_idx then E.s (E.bug "findMainLine: inf");
         match sources.(src')#getInstrInfLines idx' with 
           r :: _ -> r.rFirst
         | _ -> E.s (E.bug "findMainLine: no inf lines")
       end else begin
         (* Main is a superior line *)
         if src' = 0 then E.s (E.bug "findMainLine");
         findMainLine (src' - 1) (sources.(src')#getInstrSupLine idx')
       end
     in
     let main_line = findMainLine src idx in

     (* Make sure we are paging in the entire range *)
     pageInMainLine (main_line - pagingSize / 2 + 1);
     pageInMainLine (main_line + pagingSize / 2);

     let pageno = mainLineToPageNumber main_line in 
     let pd = 
       match GA.get pages pageno with 
         Loaded pd -> pd
       | _ -> E.s (E.bug "highlighline: page not loaded")
     in
     (* Find the window lines for the start and end *)
     let start_line, nrLines = sources.(src)#getInstrVirtualLines idx in
     
     let start_line = start_line - pd.pgVirtualLine + pd.pgWindowLine in
(*         
         ignore (E.log "selectLine: src=%d, idx=%d,start_line=%d,end_line=%d\n"
                   src idx start_line end_line);
*)
     self#highlightLine start_line nrLines;

     
   method private selectLine (src: int) (idx: int) = begin
     (* Find the window line *)
     if src < 0 || src > Array.length sources then 
       E.s (E.bug "Sourcewin: viewLine invalid source %d\n" src);

     (* Find the least superior that is shown *)
     let rec findLeastSuperior (src: int) (idx: int) = 
       if src < 0 || idx = 0 then ()
       else if not !(sources.(src)#show) then 
         findLeastSuperior (src - 1) (sources.(src)#getInstrSupLine idx)
       else begin
         self#showLine src idx
       end
     in
     findLeastSuperior src idx
         
  end


  method private highlightLine (start_line: int) (nrLines: int) = begin
     (* first delete old hilight *)
     Text.tag_delete srccode ["hilight"];
     (* now lighlight current *)
     Text.tag_add srccode 
       ~tag:"hilight" 
       ~start:(`Linechar(start_line,0),[`Linestart])
       ~stop:(`Linechar(start_line + nrLines - 1,0),[`Lineend]);
     Text.see srccode ~index:(`Linechar(start_line+10,0),[`Linestart]);
     Text.see srccode ~index:(`Linechar(start_line-2,0),[`Linestart]);
     Text.tag_configure srccode ~tag:"hilight" ~foreground:`Red
  end

  (** The instr position for the current event *)

  val mutable eventInstrPosition = 0, 0
  val mutable eventWindowLine = 0

  (* This is called from Bind in children classes *)
  method private markEvent (e: Tk.eventInfo) = 
    eventWindowLine <- 0;
    eventInstrPosition <- 0, 0;

    let (loc: text_index) = (`Atxy (e.ev_MouseX, e.ev_MouseY)) in
    let (indx: textIndex) = (loc, []) in
    let (`Linechar (wind_line, _)) = Text.index srccode indx in
    self#highlightLine wind_line 1;

    eventWindowLine <- wind_line;

    (* We must map from window line to virtual line *)
    let event_virtual_line = windowLineToVirtualLine wind_line in
    try 
      eventInstrPosition <- 
        H.find virtualLineToInstrPosition event_virtual_line
    with Not_found -> E.s (E.bug "SourceWin:markEvent")
       
end

